#!/usr/bin/env python

# pylint: disable=R0904
"""
Module with implementation of Interference Alignment (IA) algorithms.

Note that all IA algorithms require the channel object and any change to
the channel object must be performed before calling the `solve` method of
the IA algorithm object. This includes generating the channel and setting
the noise variance.
"""

import itertools
from abc import ABCMeta, abstractmethod
from copy import copy
from itertools import product
from typing import (Any, List, Optional, Sequence, Tuple, Type, TypeVar, Union,
                    cast)

import numpy as np
from scipy import optimize

from ..channels import multiuser as muchannels
from ..util.misc import (get_principal_component_matrix,
                         least_right_singular_vectors, leig, peig,
                         update_inv_sum_diag)
from .iabase import IASolverBaseClass

__all__ = [
    'AlternatingMinIASolver', 'MaxSinrIASolver', 'MinLeakageIASolver',
    'ClosedFormIASolver', 'MMSEIASolver', 'GreedStreamIASolver',
    'BruteForceStreamIASolver', 'IterativeIASolverBaseClass'
]

FloatOrFloatSequence = TypeVar("FloatOrFloatSequence", float, Sequence[float])
IntOrIntSequence = TypeVar("IntOrIntSequence", int, Sequence[int])


# xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
# xxxxxxxxxxxxxxx ClosedFormIASolver class xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
# xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
class ClosedFormIASolver(IASolverBaseClass):
    """
    Implements the closed form Interference Alignment algorithm as
    described in the paper "Interference Alignment and Degrees of Freedom
    of the K User Interference Channel [CadambeDoF2008]_".

    Parameters
    ----------
    multiUserChannel : muchannels.MultiUserChannelMatrix
        The multiuser channel.
    use_best_init : bool
        If true, all possible initializations (subsets of the
        eigenvectors of the 'E' matrix)for the first precoder will be
        tested and the best solution will be used. If false, the first
        initialization will be used.

    Notes
    -----

    .. [CadambeDoF2008] V. R. Cadambe and S. A. Jafar, "Interference
       Alignment and Degrees of Freedom of the K User Interference
       Channel," IEEE Transactions on Information Theory 54, pp. 3425-3441,
       Aug. 2008.
    """
    def __init__(self,
                 multiUserChannel: muchannels.MultiUserChannelMatrix,
                 use_best_init: bool = True):
        IASolverBaseClass.__init__(self, multiUserChannel)
        self._use_best_init = use_best_init

    # noinspection PyPep8
    def _calc_E(self) -> np.ndarray:
        """
        Calculates the "E" matrix, given by

        :math:`\\mtE = \\mtH_{31}^{-1}\\mtH_{32}\\mtH_{12}^{-1}\\mtH_{13}\\mtH_{23}^{-1}\\mtH_{21}`.

        Returns
        -------
        np.ndarray
            The "E" matrix.
        """
        # $\mtE = \mtH_{31}^{-1}\mtH_{32}\mtH_{12}^{-1}\mtH_{13}\mtH_{23}^{-1}\mtH_{21}$

        H31 = self._get_channel(2, 0)
        H32 = self._get_channel(2, 1)
        H12 = self._get_channel(0, 1)
        H13 = self._get_channel(0, 2)
        H23 = self._get_channel(1, 2)
        H21 = self._get_channel(1, 0)

        E = (np.linalg.solve(H31, H32)).dot(
            (np.linalg.solve(H12, H13)).dot(np.linalg.solve(H23, H21)))

        return E

    def _calc_all_F_initializations(self, Ns: int) -> List[np.ndarray]:
        """
        Calculates all possible initializations for the first precoder
        (self._F[0]).

        The precoder self._F[0] is initialized with a subset of the
        eigenvectors of the matrix 'E'. Therefore, this method returns all
        possible subsets of the eigenvectors of the matrix 'E'.

        Parameters
        ----------
        Ns : int
            Number of streams of the first user.

        Returns
        -------
        all_initializations : list[np.ndarray]
            All possible subsets (with size Ns) of the eigenvectors of the
            matrix E
        """
        E = self._calc_E()
        eigenvectors = np.linalg.eig(E)[1]
        num_eigenvectors = eigenvectors.shape[1]

        all_subsets = []

        for comb_index in itertools.combinations(range(num_eigenvectors), Ns):
            all_subsets.append(eigenvectors[:, comb_index])

        return all_subsets

    def _updateF(self, F0: Optional[np.ndarray] = None) -> None:
        """
        Find the precoders.

        Parameters
        ----------
        F0 : np.ndarray
            The first precoder. If not provided, the matrix 'E' will be
            calculated (with the _calc_E method) and the the first Ns
            eigenvectors will be used as F0.
        """
        self._clear_precoder_filter()

        # The number of users is always 3 for the ClosedFormIASolver class
        self._F = np.zeros(3, dtype=np.ndarray)

        if F0 is None:
            E = self._calc_E()
            Ns0 = self.Ns[0]

            # The first precoder is given by any subset of the
            # eigenvectors of the "E" matrix. We simple get the first
            # Ns_0 eigenvectors of E.
            eigenvectors = np.linalg.eig(E)[1]
            F0 = eigenvectors[:, 0:Ns0]
            self._F[0] = F0
        else:
            self._F[0] = F0

        # The second precoder is given by $\mtH_{32}^{-1}\mtH_{31}\mtV_1$
        invH32 = np.linalg.pinv(self._get_channel(2, 1))
        H31 = self._get_channel(2, 0)
        self._F[1] = np.dot(invH32, np.dot(H31, F0))
        # The third precoder is given by $\mtH_{23}^{-1}\mtH_{21}\mtV_1$
        invH23 = np.linalg.pinv(self._get_channel(1, 2))
        H21 = self._get_channel(1, 0)
        self._F[2] = np.dot(invH23, np.dot(H21, F0))

        # Normalize the precoders
        self._F[0] /= np.linalg.norm(self._F[0], 'fro')
        self._F[1] /= np.linalg.norm(self._F[1], 'fro')
        self._F[2] /= np.linalg.norm(self._F[2], 'fro')

    # noinspection PyUnresolvedReferences
    def _updateW(self) -> None:
        """Find the receive filters
        """
        self._clear_receive_filter()

        # The number of users is always 3 for the ClosedFormIASolver class
        self._W = np.zeros(3, dtype=np.ndarray)

        A0 = np.dot(self._get_channel(0, 1), self.F[1])
        self._W[0] = leig(np.dot(A0,
                                 A0.transpose().conjugate()), self.Ns[0])[0]

        A1 = np.dot(self._get_channel(1, 0), self.F[0])
        self._W[1] = leig(np.dot(A1,
                                 A1.transpose().conjugate()), self.Ns[1])[0]

        A2 = np.dot(self._get_channel(2, 0), self.F[0])
        self._W[2] = leig(np.dot(A2,
                                 A2.transpose().conjugate()), self.Ns[2])[0]

    # noinspection PyUnboundLocalVariable
    def solve(self,
              Ns: IntOrIntSequence,
              P: Optional[FloatOrFloatSequence] = None) -> None:
        """
        Find the IA solution.

        This method updates the 'F' and 'W' member variables.

        Parameters
        ----------
        Ns : int | np.ndarray
            Number of streams of each user.
        P : np.ndarray | List[float] | float, optional
            Power of each user. If not provided, a value of 1 will be used
            for each user.
        """
        assert self.K == 3, ('The ClosedFormIASolver class only works'
                             ' in a MIMO-IC scenario with 3 users.')

        if isinstance(Ns, int):
            Ns = np.ones(3, dtype=int) * Ns
        else:
            # noinspection PyTypeChecker
            assert len(Ns) == 3

        # Help the type system knowing that at this point Ns is a Sequence[int]
        assert (not isinstance(Ns, int))

        # This will create a new array so that we can modify self._Ns
        # internally without changing the original Ns variable passed to
        # the solve method.
        self._Ns = np.array(Ns)

        self.P = P

        if self._use_best_init is True:
            # xxxxx Case when the best solution should be used xxxxxxxxxxxx
            best_sum_capacity = 0
            all_initializations = self._calc_all_F_initializations(Ns[0])

            # Lambda function to calculate the sum capacity from the SINR
            # values (in linear scale)
            def calc_capacity(sinr: np.ndarray) -> float:
                return cast(float, np.sum(np.log2(1 + sinr)))

            for F0 in all_initializations:
                self._updateF(F0)
                self._updateW()

                # Calculates the sum capacity
                sinr_all_k = self.calc_SINR()

                # Array with the sum capacity of each user
                sum_capacity = map(calc_capacity, sinr_all_k)
                # Total sum capacity
                total_sum_capacity = np.sum(list(sum_capacity))

                if total_sum_capacity > best_sum_capacity:
                    best_sum_capacity = total_sum_capacity
                    best_F = self._F
                    best_W = self._W

            self._clear_precoder_filter()
            self._clear_receive_filter()
            self._F = best_F
            self._W = best_W
            # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
        else:
            # xxxxx Case when the first solution should be used xxxxxxxxxxx
            self._updateF()
            self._updateW()
            # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx


# xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
# xxxxxxxxxx Base Class for all iterative IA algorithms xxxxxxxxxxxxxxxxxxx
# xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
class IterativeIASolverBaseClass(IASolverBaseClass):
    """
    Base class for all Iterative IA algorithms.

    All subclasses of IterativeIASolverBaseClass must implement at least
    the _updateF and _updateW methods.

    Solving an iterative algorithm usually involves some initialization and
    then performing a "step" a given number of times until convergence. The
    initialization code is performed in the `_solve_init` method while the
    "step" corresponds to the `_step` method.

    The initialization code is defined here as simply initializing the
    precoder with a random matrix and then calling the `_updateW` method
    (which must be implemented in a subclass) to update the receive
    filter. This is usually what you want but any subclass can redefine
    `_solve_init` if a different initialization is required.

    The "step" part usually involves updating the precoder and then
    updating the receive filters. The definition of `_step` here calls two
    methods that MUST be defined in subclasses, the _updateF method and the
    _updateW method. If anything else is required in the "step" part then
    the _step method can be redefined in a subclass, but even in that case
    it should call the _updateF and _updateW methods instead of
    implementing everything in your redefined _step method.

    Parameters
    ----------
    multiUserChannel : muchannels.MultiUserChannelMatrix
        The multiuser channel.
    """

    # The IterativeIASolverBaseClass is an abstract class and the
    # '_updateF' and '_updateW' methods (marked as abstract) must be
    # implemented in a subclass.
    __metaclass__ = ABCMeta

    def __init__(self, multiUserChannel: muchannels.MultiUserChannelMatrix):
        IASolverBaseClass.__init__(self, multiUserChannel)

        # Count how many times the step method was called. This will be
        # reseted to zero whenever the channel or the precoder is
        # reinitialized.
        self._runned_iterations = 0
        # Number of times the step method is called in the solve method.
        self.max_iterations = 50

        # Relative change of the precoder in one iteration to the next
        # one. If the relative change from one iteration to the next one is
        # lower than this factor then the algorithm will stop the
        # iterations before the max_iterations limit is reached.
        self.relative_factor = 1e-6

        # We can use the closed form IA solver or the Alternating
        # Minimization IA solver to initialize the iterative
        # algorithm. This can reduce the number of iterations required for
        # convergence or even get better performance.
        #
        # Note that this is controlled by the 'initialize_with'
        # variable. If it is 'random' the initialization is performed with
        # random precoders. If it is 'closed_form' the initialization is
        # performed with the Closed Form algorithm, while if it is
        # 'alt_min' the initialization is performed with the Alternating
        # Minimizations algorithm.
        self._closed_form_ia_solver = ClosedFormIASolver(multiUserChannel,
                                                         use_best_init=True)

        # If self is of the class AlternatingMinIASolver, then the
        # initialization with the Alternating Minimizations algorithm makes
        # no sense.
        self._alt_min_ia_solver: Optional[AlternatingMinIASolver]
        if isinstance(self, AlternatingMinIASolver):
            self._alt_min_ia_solver = None
        else:
            self._alt_min_ia_solver = AlternatingMinIASolver(multiUserChannel)

        # Can be: 'random', 'closed_form', 'alt_min', or 'fix'
        #
        # 'random' -> Precoder will be initialized randomly and then the
        #             receive filter will be updated (using the _updateW
        #             method which is implemented in the subclass..
        #
        # 'closed_form' : -> The precoder and receive filters will be
        #                    initialized with the solution of the closed
        #                    form algorithm.
        #
        # 'alt_min' : -> The precoder and receive filters will be
        #                initialized with the solution of the alternating
        #                minimizations algorithm.
        #
        # 'fix' : The precoder is not initialized and the current value of
        #         self.F is used. This initialization type should only be
        #         used to call the 'solve' method a second time after it
        #         has been called before with another of the initialization
        #         type. This allows performing more iterations continuing
        #         from the previous solution . This is specially useful for
        #         debugging.
        self._initialize_with = 'random'

    # xxxxxxxxxx initialize_with property xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
    @property
    def initialize_with(self) -> str:
        """Get method for the initialize_with property."""
        return self._initialize_with

    @initialize_with.setter
    def initialize_with(self, value: str) -> None:
        """
        Set method for the initialize_with property.

        Parameters
        ----------
        value : str
            A string with the initialization method.
        """
        options = ['random', 'alt_min', 'closed_form', 'fix', 'svd']
        if value in options:
            self._initialize_with = value
        else:
            msg = "unknown initialization option: '{0}'".format(value)
            raise RuntimeError(msg)

    # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx

    @property
    def runned_iterations(self) -> int:
        """Get method for the runned_iterations property."""
        return self._runned_iterations

    def clear(self) -> None:
        """
        Clear the IA Solver object.

        All member attributes that are updated during the solve method,
        such as the precoder and receive filters, will be cleared. The
        other attributes that correspond to "configuration" such as the
        channel object won't be changed

        Notes
        -----
        You should overwrite this method in subclasses that pass parameters
        to the __init__ method, since here we call __init__ without
        arguments which is probably not what you want.
        """
        IASolverBaseClass.clear(self)
        self._runned_iterations = 0

    @abstractmethod
    def _updateF(self) -> None:  # pragma: no cover
        """
        Update the precoders.

        Notes
        -----
        This method should be implemented in the derived classes

        See also
        --------
        _step
        """
        raise NotImplementedError("_updateF: Not implemented")

    @abstractmethod
    def _updateW(self) -> None:  # pragma: no cover
        """
        Update the receive filters.

        Notes
        -----
        This method should be implemented in the derived classes

        See also
        --------
        _step
        """
        raise NotImplementedError("_updateW: Not implemented")

    def _step(self) -> None:
        """Performs one iteration of the algorithm.

        This method does not return anything, but instead updates the
        precoder and receive filter.
        """
        self._updateF()
        self._updateW()

    # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
    # Overwrite method in the IASolverBaseClass that change the precoder so
    # that they also reset the _runned_iterations attribute to zero.
    def randomizeF(self,
                   Ns: Union[int, List[int], Sequence[int]],
                   P: Optional[np.ndarray] = None) -> None:
        self._runned_iterations = 0
        # randomizeF in the base class will set `self._P` to `P`
        super(IterativeIASolverBaseClass, self).randomizeF(Ns, P)

    randomizeF.__doc__ = IASolverBaseClass.randomizeF.__doc__

    # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx

    def _before_initialize_W_func(self) -> None:
        """
        Method run in any of the initialize methods after the precoder is
        initialized but before the receive filter is initialized.
        """
        pass

    def _initialize_F_randomly_and_find_W(self, Ns: Sequence[int],
                                          P: np.ndarray) -> None:
        """
        Initialize the IA Solution from a random matrix.

        The implementation here simple initializes the precoder variable
        and then calculates the initial receive filter.

        Parameters
        ----------
        Ns : int np.ndarray
            Number of streams of each user.
        P : np.ndarray
            Power of each user. If not provided, a value of 1 will be used
            for each user.
        """
        self.randomizeF(Ns, P)

        # Method called before the _updateW method
        self._before_initialize_W_func()

        # Calculate the receive filters
        self._updateW()

    # noinspection PyUnusedLocal
    def _initialize_F_with_svd_and_find_W(self, Ns: IntOrIntSequence,
                                          P: np.ndarray) -> None:
        """
        Initialize the IA Solution from the most significant singular
        vectors of each user's channel.

        The implementation here simple initializes the precoder variable of
        each user as the most significant singular vector(s) of that
        user. After that, calculate the initial receive filters.

        Parameters
        ----------
        Ns : int | np.ndarray
            Number of streams of each user.
        P : np.ndarray
            Power of each user. If not provided, a value of 1 will be used
            for each user.
        """
        if isinstance(Ns, int):
            Ns = np.ones(self.K, dtype=int) * Ns

        # Help the type system knowing that at this point Ns is a Sequence[int]
        assert (not isinstance(Ns, int))

        # Create the precoder variable
        self._F = np.empty(self.K, dtype=np.ndarray)

        # Set the precoder of each user with the most significant singular
        # vector(s) of that user's channel
        for k in range(self.K):
            Hkk = self._get_channel(k, k)
            # The second variable returned by least_right_singular_vectors
            # has the corresponds to the most significant singular
            # vectors.
            _, V1, _ = least_right_singular_vectors(Hkk, self.Nr[k] - Ns[k])
            self._F[k] = V1 / np.linalg.norm(V1, 'fro')

        # Method called before the _updateW method
        self._before_initialize_W_func()

        # Calculate the receive filters
        self._updateW()

    def _dont_initialize_F_and_only_and_find_W(self, *_: Any) -> None:
        """
        Initialize the IA Solution from a random matrix.

        The implementation here simple initializes the precoder variable
        and then calculates the initial receive filter.

        Note: The `dummy1` and `dummy2` arguments have no effect. They only
        exist to keep the signature of this method equal to the signature
        of other *initialize* methods.
        """
        # The current value of self._F and self._full_F will be used.
        if self._F is None:
            msg = ("The precoder must be manually set, since you specified"
                   " the 'fix' initialize_with option.")
            raise RuntimeError(msg)

        # For the 'fix' initialization type we get the number of streams
        # from the initialized precoders.
        self._Ns = np.array([F.shape[1] for F in self._F])

        # Method called before the _updateW method
        self._before_initialize_W_func()

        self._updateW()

    def _initialize_F_and_W_from_closed_form(self, Ns: IntOrIntSequence,
                                             P: np.ndarray) -> None:
        """
        Initialize the IA Solution from the closed form IA solver.

        Parameters
        ----------
        Ns : int | np.ndarray
            Number of streams of each user.
        P : np.ndarray
            Power of each user. If not provided, a value of 1 will be used
            for each user.
        """
        # Clear all precoders and receive filters
        self._clear_precoder_filter()
        self._clear_receive_filter()

        self.P = P

        self._closed_form_ia_solver.solve(Ns, P)
        self._F = self._closed_form_ia_solver.F

        # Method called before the _updateW method
        self._before_initialize_W_func()

        self._W = self._closed_form_ia_solver.W

    def _initialize_F_and_W_from_alt_min(self, Ns: IntOrIntSequence,
                                         P: np.ndarray) -> None:
        """
        Initialize the IA Solution from the Alternating Minimizations IA
        solver.

        Parameters
        ----------
        Ns : int | np.ndarray
            Number of streams of each user.
        P : np.ndarray
            Power of each user. If not provided, a value of 1 will be used
            for each user.
        """
        # Clear all precoders and receive filters
        self._clear_precoder_filter()
        self._clear_receive_filter()

        self.P = P

        assert (self._alt_min_ia_solver is not None)
        self._alt_min_ia_solver.max_iterations = self.max_iterations
        self._alt_min_ia_solver.solve(Ns, P)

        self._F = self._alt_min_ia_solver.F

        # Method called before the _updateW method
        self._before_initialize_W_func()

        self._W = np.empty(self.K, dtype=np.ndarray)
        for k in range(self.K):
            Wk = self._alt_min_ia_solver.W[k]
            self._W[k] = Wk / np.linalg.norm(Wk, 'fro')

    def _solve_init(self, Ns: IntOrIntSequence,
                    P: FloatOrFloatSequence) -> None:
        """
        Code run in the `solve` method before the loop that run the :meth:`_step`
        method.

        The implementation here simple initializes the precoder variable
        and then calculates the initial receive filter.

        Parameters
        ----------
        Ns : int | np.ndarray
            Number of streams of each user.
        P : float | np.ndarray
            Power of each user. If not provided, a value of 1 will be used
            for each user.
        """
        self.P = P

        # initialize_with can be: 'random', 'fix', 'closed_form',
        # 'alt_min', or 'svd'
        options = {
            'random': self._initialize_F_randomly_and_find_W,
            'alt_min': self._initialize_F_and_W_from_alt_min,
            'closed_form': self._initialize_F_and_W_from_closed_form,
            'fix': self._dont_initialize_F_and_only_and_find_W,
            'svd': self._initialize_F_with_svd_and_find_W
        }

        initialzie_func = options[self.initialize_with]
        initialzie_func(Ns, P)  # type: ignore

    def _solve_finalize(self) -> None:  # pragma: no cover
        """
        Perform any post processing after the solution has been found.

        Some of the found precoders may be a singular matrix. In that case,
        we need to remove the dimensions with zero energy from both the
        found precoder and the receive filter.
        """
        # Store the index of the users from which we need to modify the
        # precoders and receive filters
        mod_users = []
        num_significant_sing_values = []
        assert (self._F is not None)
        assert (self._full_F is not None)
        for k in range(self.K):
            # We only need to perform further actions if more then one
            # streams is transmitted. In that case we need to test if the
            # final precoder has a large condition number indicating that
            # some dimensions has very low energy and could be
            # discarded. If only one stream is transmitted then the
            # condition number is always equal to 1.
            if self.Ns[k] > 1:
                [_, S, _] = np.linalg.svd(self._F[k])
                # If the condition number os large, then there is some
                # dimension with zero energy
                cond = S.max() / S.min()
                if cond > 1e4:
                    mod_users.append(k)
                    # [U, S, V_H] = np.linalg.svd(self._F[k])
                    max_sing_value = S.max().item()
                    # Calculate the number of significative singular
                    # values. Basically, any singular value (and
                    # corresponding dimension) lower then
                    # max_sing_value/1e4 will be discarded.
                    n = np.count_nonzero(np.greater(S, max_sing_value / 1.0e4))

                    # Store the number of significative singular values for
                    # that user
                    num_significant_sing_values.append(n)

                    new_F = get_principal_component_matrix(self._F[k], n)

                    # Normalize new_F
                    new_F /= np.linalg.norm(new_F, 'fro')

                    self._F[k] = new_F

                    if self._full_F[k] is not None:
                        # Original norm of the _full_F[k] precoder
                        original_norm = np.linalg.norm(self._full_F[k], 'fro')
                        new_full_F = get_principal_component_matrix(
                            self._full_F[k], n)
                        # Restore the original norm
                        new_full_F = new_full_F / np.linalg.norm(
                            new_full_F, 'fro') * original_norm
                        self._full_F[k] = new_full_F

                    self.Ns[k] = n

        # If we modified any of the precoders then the mod_users list has
        # the index of the users whose precoders were modified. We need to
        # also modify the receive filter for those users.
        if len(mod_users) > 0:
            # Note that we still need to remove the dead dimensions of
            # the receive filter. However, depending on the algorithm,
            # either the _W or the _W_H member variable was set while
            # the other is None (at this point).
            if self._W_H is None:
                assert (self._W is not None)
                # Since _W_H is None that means that we need to modify
                # the _W member variable
                for k, n in zip(mod_users, num_significant_sing_values):
                    new_W = get_principal_component_matrix(self._W[k], n)
                    self._W[k] = new_W

            elif self._W is None:
                # Since _W is None that means that we need to modify the
                #  _W_H member variable
                for k, n in zip(mod_users, num_significant_sing_values):
                    W = self._W_H[k].conj().T
                    new_W = get_principal_component_matrix(W, n)
                    self._W_H[k] = new_W.conj().T
            else:
                # If both self._W and self._W_H are not None then
                # something wrong happened. Maybe you called the self.W
                # or the self.W_H properties by mistake before
                # _solve_finalize is called (in the solve method).
                raise Exception("I should not be here.")

    @classmethod
    def _is_diff_significant(cls, F_old: np.ndarray, F_new: np.ndarray,
                             relative_factor: float) -> bool:
        """
        Test if there was any significant change from `F_old` to `F_new`.

        This method is used internally in the solve method of the
        IterativeIASolverBaseClass to detect when the precoder of a given
        iteration didn't change significantly from one iteration to
        another. This is used to stop the iterations of the algorithm and
        avoid unnecessary computations.

        Parameters
        ----------
        F_old : np.ndarray
            The precoder of all users (in a previous iteration). This is a 1D
            numpy array of numpy arrays.
        F_new : np.ndarray
            The precoder of all users (in the current iteration). This is a 1D
            numpy array of numpy arrays.
        relative_factor : float
            Relative change of the precoder in one iteration to the next one.
            If the relative change from one iteration to the next one is lower
            than this factor then the algorithm will stop the iterations before
            the max_iterations limit is reached.

        Returns
        -------
        out : bool
            True if the difference is significant, False otherwise.

        Notes
        -----
        A difference is considered significant if it is larger then 1/1000
        of the minimum value in the precoder.
        """
        K = F_old.size
        for k in range(K):
            Fk_old = F_old[k]
            Fk_new = F_new[k]
            min_value = np.abs(Fk_new).min()
            diff = np.abs(Fk_new - Fk_old)
            max_diff = diff.max()
            if max_diff > (min_value * relative_factor):
                return True

        return False

    def solve(self,
              Ns: IntOrIntSequence,
              P: Optional[FloatOrFloatSequence] = None) -> int:
        """
        Find the IA solution by performing the :meth:`_step` method several times.

        The number of times the :meth:`_step` method is run is controlled by the
        max_iterations member variable.

        Before calling the :meth:`_step` method for the first time the
        `_solve_init` method is called to perform any required
        initializations. Since iterative IA algorithms usually starts with
        a random precoder then the `_solve_init` implementation in
        IterativeIASolverBaseClass calls randomizeF.

        Parameters
        ----------
        Ns : int | np.ndarray
            Number of streams of each user.
        P : np.ndarray | List[float] | float, optional
            Power of each user. If not provided, a value of 1 will be used
            for each user.

        Returns
        -------
        Number of iterations that the iterative interference alignment
        algorithm run.

        Notes
        -----
        You probably should not overwrite this method in sub-classes of the
        IterativeIASolverBaseClass. If you want to change the
        initialization of the algorithm overwrite the _solve_init method
        instead.

        Subclasses must implement the _updateF() and _updateW() methods. If
        something else besides calling these two methods is required in the
        "step" part of the algorithm, then reimplement also the _step()
        method.
        """
        if isinstance(Ns, int):
            Ns = np.ones(self.K, dtype=int) * Ns
        else:
            # noinspection PyTypeChecker
            assert len(Ns) == self.K

        # This will create a new array so that we can modify self._Ns
        # internally without changing the original Ns variable passed to
        # the randomizeF method.
        self._Ns = np.array(Ns, dtype=int)
        self._solve_init(Ns, P)  # type: ignore

        # This will be used to detect of the precoder did not
        # significative change
        # TODO: Should I use full_F instead of F???
        old_F = self._F
        for _ in range(self.max_iterations):
            self._runned_iterations += 1
            self._step()

            # Stop the iteration earlier if the precoder does not change
            # too much
            if self._is_diff_significant(old_F, self._F,
                                         self.relative_factor) is False:
                break  # pragma: no cover

            old_F = self._F

        # Perform any post processing after the precoder and receive
        # filters where found. One possible usage for this method is to
        # remove dimensions of the precoder (and receive filter) that have
        # actually zero energy. That is, if the precoder ends up being a
        # singular matrix we can implement _solve_finalize to remove the
        # dimensions that do not contribute.
        self._solve_finalize()

        # Return the number of iterations the algorithm run
        return self._runned_iterations


# xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
# xxxxxxxxxx AlternatingMinIASolver Class xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
# xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
class AlternatingMinIASolver(IterativeIASolverBaseClass):
    """
    Implements the "Interference Alignment via Alternating Minimization"
    algorithm from the paper with the same name [PetersHeathAltMin2009]_.

    This algorithm is applicable to a "K-user" scenario and it is very
    flexible in the sense that you can change the number of transmit
    antennas, receive antennas and streams per user, as well as the number
    of users involved in the IA process. However, note that alignment is
    only feasible for some cases configurations.

    An example of a common scenario is a scenario with 3 pairs or
    transmitter/receiver with 2 antennas in each node and 1 stream
    transmitted per node.

    You can determine the scenario of an AlternatingMinIASolver object by
    inferring the variables K, Nt, Nr and Ns.

    Parameters
    ----------
    multiUserChannel : muchannels.MultiUserChannelMatrix
        The multiuser channel.

    Notes
    -----

    .. [PetersHeathAltMin2009] Peters, S.W.; Heath, R.W., "Interference
       alignment via alternating minimization," Acoustics, Speech and
       Signal Processing, 2009. ICASSP 2009. IEEE International Conference
       on, pp.2445,2448, 19-24 April 2009
    """
    def __init__(self, multiUserChannel: muchannels.MultiUserChannelMatrix):
        IterativeIASolverBaseClass.__init__(self, multiUserChannel)

        self._C: List[np.ndarray] = [
        ]  # Basis of the interference subspace for each user

    # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx

    # Overwrite the set property for 'initialize_with' to remove the
    # 'alt_min' initialization type, since it does not make sense.
    # noinspection PyMethodOverriding,PyIncorrectDocstring
    @IterativeIASolverBaseClass.initialize_with.setter  # type: ignore
    def initialize_with(self, value: str) -> None:
        """Set method for the initialize_with property."""
        if value == 'alt_min':  # pragma: nocover
            msg = "Can't use '{0}' initialization with '{1}' class '{0}'"
            raise RuntimeError(msg.format(value, self.__class__.__name__))

        IterativeIASolverBaseClass.initialize_with.fset(  # type: ignore
            self, value)

    def get_cost(self) -> float:
        """
        Get the Cost of the algorithm for the current iteration of the
        precoder.

        Returns
        -------
        cost : float
            The Cost of the algorithm for the current iteration of the
            precoder. This is a real non-negative number.
        """
        Cost = 0
        # This will get all combinations of (k,l) without repetition. This
        # is equivalent to two nested for loops with an if statement to
        # only execute the code only when `k` is different of `l`.
        all_kl_indexes = itertools.permutations(range(self.K), 2)

        for kl in all_kl_indexes:
            (k, l) = kl
            Hkl_Fl = np.dot(self._get_channel(k, l), self.full_F[l])
            Cost += np.linalg.norm(
                Hkl_Fl -
                np.dot(np.dot(self._C[k], self._C[k].transpose().conjugate()),
                       Hkl_Fl), 'fro')**2

        return Cost

    def _before_initialize_W_func(self) -> None:
        """
        Method run in any of the initialize methods after the precoder is
        initialized but before the receive filter is initialized.
        """
        self._updateC()

    def _step(self) -> None:
        """
        Performs one iteration of the algorithm.

        The step method is usually all you need to call to perform an
        iteration of the Alternating Minimization algorithm. It will update
        C, then update F and finally update W.

        See also
        --------
        _updateC, _updateF, _updateW
        """
        # Note that before the `_step` method is called the first time, the
        # _solve_init method must be called. It will initialize the
        # precoders, the C matrices and the receive filters.

        self._updateF()  # Depend on the value of C
        self._updateC()  # The value of C depend on the precoders F

        # Note that in the Alternating Minimizations algorithm we do not
        # need to update the receive filters at each iteration, since the
        # precoders only depend on the values of the 'C' matrices. Because
        # of that, we will only update the receive filters in the
        # _solve_finalize method, which is called after the algorithm
        # converged.

    def _solve_finalize(self) -> None:  # pragma: no cover
        """Perform any post processing after the solution has been found.
        """
        # Note that in the Alternating Minimizations algorithm we do not
        # need to update the receive filters at each iteration, since the
        # precoders only depend on the values of the 'C' matrices. Because
        # of that, we will only update the receive filters in the
        # _solve_finalize method, which is called only once after the
        # algorithm converged.
        self._updateW()  # Depend on the value of C
        IterativeIASolverBaseClass._solve_finalize(self)

    # noinspection PyPep8
    def _updateC(self) -> None:
        """Update the value of Ck for all K users.

        Ck contains the orthogonal basis of the interference subspace of
        user k. It corresponds to the Nk-Sk dominant eigenvectors of

            :math:`\\sum_{l \\neq k} \\mtH_{k,l} \\mtF_l \\mtF_l^H \\mtH_{k,l}^H`.

        Notes
        -----
        This method is called in the :meth:`_step` method.

        See also
        --------
        _step
        """
        # $$\sum_{l \neq k} \mtH_{k,l} \mtF_l \mtF_l^H \mtH_{k,l}^H$$

        # xxxxxxxxxx New Implementation using calc_Q xxxxxxxxxxxxxxxxxxxxxx
        Ni = self.Nr - self.Ns  # Ni: Dimension of the interference subspace

        self._C = np.empty(self.K, dtype=np.ndarray)

        for k in np.arange(self.K):
            # TODO: Implement and test with external interference
            # # We are inside only of the first for loop
            # # Add the external interference contribution
            # self._C[k] = self.calc_Q(k) + self.Rk[k]

            # C[k] will receive the Ni most dominant eigenvectors of C[k]
            self._C[k] = peig(self.calc_Q(k), Ni[k])[0]
        # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx

    def _updateF(self) -> None:
        """
        Update the value of the precoder of all K users.

        Fl, the precoder of the l-th user, tries avoid as much as possible
        to send energy into the desired signal subspace of the other
        users. Fl contains the Sl least dominant eigenvectors of
        :math:`\\sum_{k \\neq l} H_{k,l}^H (I - C_k C_k^H)H_{k,l}`

        Notes
        -----
        This method is called in the :meth:`_step` method.

        See also
        --------
        _step
        """
        # $\sum_{k \neq l} \mtH_{k,l}^H (\mtI - \mtC_k \mtC_k^H)\mtH_{k,l}$

        # xxxxx Calculates the temporary variable Y[k] for all k xxxxxxxxxx
        # Note that $\mtY[k] = (\mtI - \mtC_k \mtC_k^H)$

        self._clear_precoder_filter()

        # The number of users is always 3 for the ClosedFormIASolver class
        self._F = np.zeros(self.K, dtype=np.ndarray)

        def calc_Y(Nr: int, C: np.ndarray) -> np.ndarray:
            return (np.eye(Nr, dtype=complex) -
                    np.dot(C,
                           C.conjugate().transpose()))

        Y = list(map(calc_Y, self.Nr, self._C))

        newF = np.zeros(self.K, dtype=np.ndarray)
        # This will get all combinations of (l,k) without repetition. This
        # is equivalent to two nested for loops with an if statement to
        # only execute the code only when `l` is different of `k`
        all_lk_indexes = itertools.permutations(range(self.K), 2)

        # This code will store in newF[l] the equivalent of
        # $\sum_{k \neq l} \mtH_{k,l}^H (\mtI - \mtC_k \mtC_k^H)H_{k,l}$
        for lk in all_lk_indexes:
            (l, k) = lk
            lH = self._get_channel(k, l)
            newF[l] = newF[l] + np.dot(
                np.dot(lH.conjugate().transpose(), Y[k]), lH)

        # Every element in newF is a matrix. We want to replace each
        # element by the least dominant eigenvectors of that element.
        for k in range(self.K):
            self._F[k] = leig(newF[k], self.Ns[k])[0]
            self._F[k] /= np.linalg.norm(self._F[k], 'fro')

    def _updateW(self) -> None:
        """
        Update the zero-forcing filters.

        The zero-forcing filter is calculated in the paper "MIMO
        Interference Alignment Over Correlated Channels with Imperfect
        CSI".

        Notes
        -----
        This method is called in the :meth:`_step` method.

        See also
        --------
        _step
        """
        self._clear_receive_filter()

        # Note that the formula for the receive filter in the "Interference
        # Alignment via Alternating Minimization" paper actually calculates
        # W_H instead of W.
        newW_H = np.zeros(self.K, dtype=np.ndarray)
        assert (self._F is not None)
        for k in np.arange(self.K):
            tildeHi = np.hstack(
                [np.dot(self._get_channel(k, k), self._F[k]), self._C[k]])
            newW_H[k] = np.linalg.inv(tildeHi)
            # We only want the first Ns[k] lines
            newW_H[k] = newW_H[k][0:self.Ns[k]]
        self._W_H = newW_H


# xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
# xxxxxxxxxxxxxxx MinLeakageIASolver class xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
# xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
class MinLeakageIASolver(IterativeIASolverBaseClass):
    """
    Implements the Minimum Leakage Interference Alignment algorithm as
    described in the paper "Approaching the Capacity of Wireless Networks
    through Distributed Interference Alignment [Cadambe2008]_".

    Parameters
    ----------
    multiUserChannel : muchannels.MultiUserChannelMatrix
        The multiuser channel.
    """
    def __init__(self, multiUserChannel: muchannels.MultiUserChannelMatrix):
        IterativeIASolverBaseClass.__init__(self, multiUserChannel)

    def get_cost(self) -> float:
        """
        Get the Cost of the algorithm for the current iteration of the
        precoder.

        For the Minimum Leakage Interference Alignment algorithm the cost
        is equivalent to the sum of the interference that all users see
        after applying the receive filter. That is,

            :math:`C = Tr[\\mtU_k^H \\mtQ_k \\mtU_k]`

        Returns
        -------
        cost : float
            The Cost of the algorithm for the current iteration of the
            precoder. This is a (real non-negative number).
        """
        # $$C = Tr[\mtU_k^H \mtQ_k \mtU_k]$$
        assert (self._W is not None)
        cost = 0
        for k in range(self.K):
            Qk = self.calc_Q(k)
            Wk = self._W[k]
            aux = np.dot(np.dot(Wk.transpose().conjugate(), Qk), Wk)
            cost = cost + np.trace(np.abs(aux))
        return cost

    def _calc_Uk_all_k(self) -> np.ndarray:
        """
        Calculates the receive filter of all users.

        Returns
        -------
        np.ndarray
            The Uk array of each user. This is a 1D numpy array of numpy
            arrays.
        """
        Uk = np.empty(self.K, dtype=np.ndarray)

        for k in range(self.K):
            Qk = self.calc_Q(k)
            [V, _] = leig(Qk, self.Ns[k])
            Uk[k] = V
        return Uk

    def _calc_Uk_all_k_rev(self) -> np.ndarray:
        """
        Calculates the receive filter of all users in the reverse network.

        Returns
        -------
        np.ndarray
            The Uk array of each user. This is a 1D numpy array of numpy
            arrays.
        """
        Uk_rev = np.empty(self.K, dtype=np.ndarray)
        for k in range(self.K):
            Qk_rev = self.calc_Q_rev(k)
            [V, _] = leig(Qk_rev, self.Ns[k])
            Uk_rev[k] = V
        return Uk_rev

    def _updateF(self) -> None:
        """
        Update the precoders.

        Notes
        -----
        This method is called in the :meth:`IterativeIASolverBaseClass._step` method.

        See also
        --------
        IterativeIASolverBaseClass._step

        """
        self._clear_precoder_filter()
        self._F = self._calc_Uk_all_k_rev()

    def _updateW(self) -> None:
        """
        Update the receive filters.

        Notes
        -----
        This method is called in the :meth:`IterativeIASolverBaseClass._step` method.

        See also
        --------
        IterativeIASolverBaseClass._step
        """
        self._clear_receive_filter()
        self._W = self._calc_Uk_all_k()


# xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
# xxxxxxxxxxxxxxx MaxSinrIASolver class xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
# xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
class MaxSinrIASolver(IterativeIASolverBaseClass):
    """Implements the "Interference Alignment via Max SINR" algorithm.

    This algorithm is applicable to a "K-user" scenario and it is
    described in [Cadambe2008]_.

    An example of a common scenario is a scenario with 3 pairs or
    transmitter/receiver with 2 antennas in each node and 1 stream
    transmitted per node.

    You can determine the scenario of an MaxSinrIASolver object by
    inferring the variables K, Nt, Nr and Ns.

    Parameters
    ----------
    multiUserChannel : muchannels.MultiUserChannelMatrix
        The multiuser channel.
    """
    def __init__(self, multiUserChannel: muchannels.MultiUserChannelMatrix):
        IterativeIASolverBaseClass.__init__(self, multiUserChannel)

    # noinspection PyPep8
    def _calc_Bkl_cov_matrix_first_part_rev(self, k: int) -> np.ndarray:
        """
        Calculates the first part in the equation of the Blk covariance
        matrix of the reverse channel.

        Parameters
        ----------
        k : int
            Index of the desired user.

        Returns
        -------
        Bkl_first_part_rev : np.ndarray
            First part in equation (28) of [Cadambe2008]_, but for the
            reverse channel.

        See also
        --------
        _calc_Bkl_cov_matrix_first_part_rev

        """
        # $$\sum_{j=1}^{K} \frac{P^{[j]}}{d^{[j]}} \sum_{d=1}^{d^{[j]}} \mtH^{[kj]}\mtV_{\star d}^{[j]} \mtV_{\star d}^{[j]\dagger} \mtH^{[kj]\dagger}$$
        P = self.P
        first_part = 0.0

        assert (self._W is not None)
        for j in range(self.K):
            Hkj = self._get_channel_rev(k, j)
            Hkj_H = Hkj.conjugate().transpose()
            Vj = self._W[j]

            # The lets make sure the receive filter norm is equal to one so
            # that we can correctly scale it to the desired power.
            assert np.linalg.norm(Vj, 'fro') - 1.0 < 1e-6
            assert (isinstance(P, np.ndarray))
            assert (isinstance(self._Ns, np.ndarray))
            Vj_H = Vj.conjugate().transpose()
            first_part += (float(P[j]) / self._Ns[j]) * np.dot(
                Hkj, np.dot(np.dot(Vj, Vj_H), Hkj_H))

        return first_part

    # noinspection PyPep8,PyPep8
    def _calc_Bkl_cov_matrix_second_part_rev(self, k: int,
                                             l: int) -> np.ndarray:
        """
        Calculates the second part in the equation of the Blk covariance
        matrix of the reverse channel..

        The second part is given by

            :math:`\\frac{P^{[k]}}{d^{[k]}} \\mtH^{[kk]} \\mtV_{\\star l}^{[k]} \\mtV_{\\star l}^{[k]\\dagger} \\mtH^{[kk]\\dagger}`

        Parameters
        ----------
        k : int
            Index of the desired user.
        l : int
            Index of the desired stream.

        Returns
        -------
        second_part : np.ndarray
            Second part in equation (28) of [Cadambe2008]_.
        """
        # $$\frac{P^{[k]}}{d^{[k]}} \mtH^{[kk]} \mtV_{\star l}^{[k]} \mtV_{\star l}^{[k]\dagger} \mtH^{[kk]\dagger}$$
        P = self.P

        Hkk = self._get_channel_rev(k, k)
        Hkk_H = Hkk.transpose().conjugate()

        assert (self._W is not None)
        Vkl = self._W[k][:, l:l + 1]
        Vkl_H = Vkl.transpose().conjugate()
        second_part = np.dot(Hkk, np.dot(np.dot(Vkl, Vkl_H), Hkk_H))
        assert (isinstance(P, np.ndarray))
        assert (isinstance(self._Ns, np.ndarray))
        return second_part * (float(P[k]) / self._Ns[k])

    # noinspection PyPep8
    def _calc_Bkl_cov_matrix_all_l_rev(self, k: int) -> np.ndarray:
        """
        Calculates the interference-plus-noise covariance matrix for all
        streams at "receiver" :math:`k` for the reverse channel.

        Parameters
        ----------
        k : int
            Index of the desired user.

        Returns
        -------
        Bkl_rev : np.ndarray
            Covariance matrix of all streams of user k. Each element of the
            returned 1D numpy array is a 2D numpy complex array
            corresponding to the covariance matrix of one stream of user k.
        """
        # $$\mtB^{[kl]} = \sum_{j=1}^{K} \frac{P^{[j]}}{d^{[j]}} \sum_{d=1}^{d^{[j]}} \mtH^{[kj]}\mtV_{\star l}^{[j]} \mtV_{\star l}^{[j]\dagger} \mtH^{[kj]\dagger} - \frac{P^{[k]}}{d^{[k]}} \mtH^{[kk]} \mtV_{\star l}^{[k]} \mtV_{\star l}^{[k]\dagger} \mtH^{[kk]\dagger} + \mtI_{N^{[k]}}$$
        assert (isinstance(self._Ns, np.ndarray))
        Bkl_all_l_rev = np.empty(self._Ns[k], dtype=np.ndarray)
        first_part = self._calc_Bkl_cov_matrix_first_part_rev(k)

        for l in range(self._Ns[k]):
            second_part = self._calc_Bkl_cov_matrix_second_part_rev(k, l)
            Bkl_all_l_rev[l] = first_part - second_part + (self.noise_var *
                                                           np.eye(self.Nt[k]))

        return Bkl_all_l_rev

    @classmethod
    def _calc_Ukl(cls, Hkk: np.ndarray, Vk: np.ndarray, Bkl: np.ndarray,
                  l: int) -> np.ndarray:
        """
        Calculates the Ukl matrix in equation (29) of [Cadambe2008]_.

        Parameters
        ----------
        Hkk : np.ndarray
            Channel from transmitter K to receiver K.
        Vk : np.ndarray
            Precoder of user k.
        Bkl : np.ndarray
            The previously calculates Bkl matrix in equation (28) of
            [Cadambe2008]_
        l : int
            Index of the desired stream

        Returns
        -------
        Ukl : np.ndarray
            The calculated Ukl matrix. This is a 2D numpy array (with
            self.Nr[k] rows and a single column).
        """
        Vkl = Vk[:, l:l + 1]

        Ukl = np.linalg.solve(Bkl, np.dot(Hkk, Vkl))

        Ukl /= np.linalg.norm(Ukl, 'fro')
        return Ukl

    @classmethod
    def _calc_Uk(cls, Hkk: np.ndarray, Vk: np.ndarray,
                 Bkl_all_l: np.ndarray) -> np.ndarray:
        """
        Similar to the :meth:`_calc_Ukl` method, but while :meth:`_calc_Ukl`
        calculates the receive filter (a vector) only for the :math:`l`-th
        stream :meth:`_calc_Uk` calculates a receive filter (a matrix) for all
        streams.

        Parameters
        ----------
        Hkk : np.ndarray
            Channel from transmitter K to receiver K.
        Vk : np.ndarray
            Precoder of user k.
        Bkl_all_l : np.ndarray
            Covariance matrix of all streams of user k. Each element of the
            returned 1D numpy array is a 2D numpy complex array
            corresponding to the covariance matrix of one stream of user k.

        Returns
        -------
        Uk : np.ndarray
            The receive filter for all streams of user k.
        """
        num_streams = Bkl_all_l.size
        num_Rx = Bkl_all_l[0].shape[0]
        Uk = np.zeros([num_Rx, num_streams], dtype=complex)
        for l in range(num_streams):
            Uk[:, l] = MaxSinrIASolver._calc_Ukl(Hkk, Vk, Bkl_all_l[l], l)[:,
                                                                           0]

        return Uk / np.linalg.norm(Uk, 'fro')

    def _calc_Uk_all_k(self) -> np.ndarray:
        """
        Calculates the receive filter of all users.

        Returns
        -------
        np.ndarray
            The receive filter of all users. This is a numpy array of numpy
            arrays.
        """
        Uk = np.empty(self.K, dtype=np.ndarray)
        assert (self._F is not None)
        for k in range(self.K):
            Hkk = self._get_channel(k, k)
            Bkl_all_l = self._calc_Bkl_cov_matrix_all_l(k, self.noise_var)
            Uk[k] = self._calc_Uk(Hkk, self._F[k], Bkl_all_l)
        return Uk

    def _calc_Uk_all_k_rev(self) -> np.ndarray:
        """
        Calculates the receive filter of all users for the reverse channel.

         Returns
        -------
        np.ndarray
            The receive filter of all users for the reverse channel. This is
            a numpy array of numpy arrays.
        """
        Uk = np.empty(self.K, dtype=np.ndarray)
        assert (self._W is not None)
        F = self._W  # The precoder is the receive filter of the direct channel
        for k in range(self.K):
            Hkk = self._get_channel_rev(k, k)
            Bkl_all_l = self._calc_Bkl_cov_matrix_all_l_rev(k)
            Uk[k] = self._calc_Uk(Hkk, F[k], Bkl_all_l)
        return Uk

    def _updateF(self) -> None:
        """
        Update the precoders.

        Notes
        -----
        This method is called in the :meth:`IterativeIASolverBaseClass._step` method.

        See also
        --------
        IterativeIASolverBaseClass._step
        """
        self._clear_precoder_filter()
        self._F = self._calc_Uk_all_k_rev()

    def _updateW(self) -> None:
        """
        Update the receive filters.

        Notes
        -----
        This method is called in the :meth:`IterativeIASolverBaseClass._step` method.

        See also
        --------
        IterativeIASolverBaseClass._step
        """
        self._clear_receive_filter()
        self._W = self._calc_Uk_all_k()


# xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
# xxxxxxxxxxxxxxx MMSEIASolver Class xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
# xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
class MMSEIASolver(IterativeIASolverBaseClass):
    """
    Implements the MMSE based Interference Alignment algorithm.

    This algorithm is applicable to a "K-user" scenario and it is
    described in [Peters2011]_.

    An example of a common scenario is a scenario with 3 pairs or
    transmitter/receiver with 2 antennas in each node and 1 stream
    transmitted per node.

    You can determine the scenario of an MMSEIASolver object by
    inferring the variables K, Nt, Nr and Ns.

    Parameters
    ----------
    multiUserChannel : muchannels.MultiUserChannelMatrix
        The multiuser channel.

    Notes
    -----

    .. [Peters2011] S. W. Peters and R. W. Heath, "Cooperative Algorithms
       for MIMO Interference Channels," vol. 60, no. 1, pp. 206-218, 2011.
    """
    def __init__(self, multiUserChannel: muchannels.MultiUserChannelMatrix):
        IterativeIASolverBaseClass.__init__(self, multiUserChannel)

        self._mu: Optional[np.ndarray] = None

    def _solve_init(self, Ns: IntOrIntSequence, P: np.ndarray) -> None:
        """
        Code run in the `solve` method before the loop that run the
        :meth:`IterativeIASolverBaseClass._step` method.

        The implementation here simple initializes the precoder variable
        and then calculates the initial receive filter.

        Parameters
        ----------
        Ns : int | np.ndarray
            Number of streams of each user.
        P : np.ndarray
            Power of each user. If not provided, a value of 1 will be used
            for each user.
        """
        IterativeIASolverBaseClass._solve_init(self, Ns, P)
        self._mu = np.zeros(self.K, dtype=float)

    # noinspection PyPep8
    def _calc_Uk(self, k: int) -> np.ndarray:
        """
        Calculates the receive filter of the k-th user.

        Parameters
        ----------
        k : int
            User index

        Returns
        -------
        Uk : np.ndarray
            The receive filter of the user 'k'.
        """
        # $$\mtU_k = \left( \sum_{i=1}^K \mtH_{ki} \mtV_i \mtV_i^H \mtH_{ki}^H + \sigma_n^2 \mtI \right)^{-1} \mtH_{kk} \mtV_k$$
        Hkk = self._get_channel(k, k)
        Vk = self.full_F[k]

        sum_term = 0
        for i in range(self.K):
            Hki = self._get_channel(k, i)
            Vi = self.full_F[i]
            aux = np.dot(Hki, Vi)
            sum_term = sum_term + np.dot(aux, aux.conj().T)

        sum_term2 = sum_term + self.noise_var * np.eye(self.Nr[k])
        ":type: np.ndarray"

        Uk = np.linalg.solve(sum_term2, np.dot(Hkk, Vk))
        return Uk

    def _updateW(self) -> None:
        """
        Updates the receive filter of all users.
        """
        new_W = np.zeros(self.K, dtype=np.ndarray)
        for k in range(self.K):
            new_W[k] = self._calc_Uk(k)

        self._clear_receive_filter()
        self._W = new_W

    @staticmethod
    def _calc_Vi_for_a_given_mu(sum_term: np.ndarray, mu_i: float,
                                H_herm_U: np.ndarray) -> np.ndarray:
        """
        Calculates the value of Vi for the given parameters.

        This method is called inside _calc_Vi.

        Parameters
        ----------
        sum_term : np.ndarray
            The summation term in the formula to calculate the precoder.
        mu_i : float
            The value of the lagrange multiplier
        H_herm_U : np.ndarray
            The value of :math:`H_ii^H U_i`

        Returns
        -------
        np.ndarray
            The Vi matrix for the given parameters.
        """
        N = sum_term.shape[0]
        Vi = np.linalg.solve(sum_term + mu_i * np.eye(N), H_herm_U)
        # Vi = np.dot(np.linalg.inv(sum_term + mu_i * np.eye(N)),
        #             H_herm_U)

        return Vi

    @staticmethod
    def _calc_Vi_for_a_given_mu2(inv_sum_term: np.ndarray, mu_i: float,
                                 H_herm_U: np.ndarray) -> np.ndarray:
        """
        Calculates the value of Vi for the given parameters.

        This method is called inside _calc_Vi.

        Parameters
        ----------
        inv_sum_term : np.ndarray
            The inverse of the summation term in the formula to calculate
            the precoder when mu_i is equal to zero.
        mu_i : float
            The value of the lagrange multiplier
        H_herm_U : np.ndarray
            The value of :math:`H_ii^H U_i`

        Returns
        -------
        np.ndarray
            The Vi matrix for the given parameters.
        """
        N = inv_sum_term.shape[0]
        diagonal = mu_i * np.ones(N)  # Vector of N elements
        new_inv = update_inv_sum_diag(inv_sum_term, diagonal)
        Vi = np.dot(new_inv, H_herm_U)
        return Vi

    # noinspection PyPep8
    def _calc_Vi(self, i: int, mu_i: Optional[float] = None) -> np.ndarray:
        """
        Calculates the precoder of the i-th user.

        Parameters
        ----------
        i : int
            User index
        mu_i : float, optional
            The value of the Lagrange multiplier. If it is None (default),
            then the best value will be found and used to calculate the
            precoder.

        Returns
        -------
        Vi : np.ndarray
            The calculate precoder of the i-th user.
        """
        # $$\mtV_i = \left( \sum_{k=1}^K \mtH_{ki}^H \mtU_k \mtU_k^H \mtH_{ki} + \mu_i \mtI \right)^{-1} \mtH_{ii}^H \mtU_i$$

        # H_{ii}^H * Ui
        Hii_herm_U = np.dot(self._get_channel(i, i).conj().T, self.W[i])

        # xxxxx Calculates the Summation term xxxxxxxxxxxxxxxxxxxxxxxxxxxxx
        sum_term = np.array([0.0])
        for k in range(self.K):
            # H_{ki}^H * Uk
            aux = np.dot(self._get_channel(k, i).conj().T, self.W[k])
            sum_term = sum_term + np.dot(aux, aux.conj().T)
        # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx

        # xxxxxxxxxx Perform diagonal loading if it is necessary xxxxxxxxxx
        # Occasionally (specially for the low SNR values) sum_term may be a
        # singular matrix. In that case, we add a diagonal loading factor
        # to make the sum_term non-singular. This diagonal loading
        # corresponding to summing an identity matrix times a load_factor
        # to the sum_term, where this load_factor is arbitrarily calculated
        # as 1/100 of the mean of the eigen values of the singular
        # sum_term.

        # Calculates the SVD of sum_term so that we can calculate the
        # condition number.
        [_, S, _] = np.linalg.svd(sum_term)
        cond = S.max() / S.min()
        load_factor = 0.0
        # If the condition number is larger than 1e8 we consider sum_term
        # as a singular matrix, which means that we will perform the
        # diagonal loading
        if cond > 5e4:  # pragma: no cover
            # Calculates the load_factor (arbitrarily chosen as 1/100 the
            # mean of the current singular values of sum_term).
            load_factor = S.mean() / 100.0
            # pylint: disable= E1103
            sum_term += np.eye(sum_term.shape[0]) * load_factor
        # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx

        # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
        # xxxxxxxxxx Case when the best mu value must be found xxxxxxxxxxxx
        # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
        assert (isinstance(self._mu, np.ndarray))
        if mu_i is None:
            # xxxxx Define the function that will be optimized xxxxxxxxxxxx
            # noinspection PyShadowingNames,PyIncorrectDocstring
            def func(local_mu: float, local_sum_term: np.ndarray,
                     local_Hii_herm_U: np.ndarray, local_P: float) -> float:
                """
                Function that will be optimized to find the best value of
                local_mu.
                """
                Vi = self._calc_Vi_for_a_given_mu(local_sum_term, local_mu,
                                                  local_Hii_herm_U)
                norm = np.linalg.norm(Vi, 'fro')
                cost = (norm**2) - local_P
                return cast(float, cost)

            # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx

            # xxxxxxxxxx Scale Hii_herm_U and sum_term xxxxxxxxxxxxxxxxxxxx
            # Depending of the transmit power the value of 'Hii^H Ui' and
            # `sum_term` can be very big (or maybe very small). This may
            # cause precision problems in the optimization of the Lagrange
            # multiplier mu. To avoid this problem we will scale both
            # 'Hii^H Ui' and `sum_term` by this `scale_factor` to avoid
            # that.
            #
            # Note that THIS WILL NOT CHANGE the final value of Vi,
            # although the Lagrange multiplier will be different.
            scale_factor = np.linalg.norm(Hii_herm_U)
            Hii_herm_U /= scale_factor
            sum_term /= scale_factor
            # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx

            min_mu_i = 0.0

            cost = func(min_mu_i, sum_term, Hii_herm_U, self.P[i])
            # If cost is lower than or equal to zero then the power
            # constraint is already satisfied and we are done. The value of
            # mu will be min_mu_i.
            if cost <= 0:
                assert (self._mu is not None)
                mu_i = min_mu_i
                Vi = self._calc_Vi_for_a_given_mu(sum_term, mu_i, Hii_herm_U)
                self._mu[i] = mu_i
            else:
                try:
                    # If we are not done yet then we need to perform the
                    # bisection method to find the best mu value between
                    # min_mu_i and max_mu_i
                    mu_i = optimize.newton(  # pylint: disable= E1101
                        func,
                        min_mu_i,
                        args=(sum_term, Hii_herm_U, self.P[i]),
                        maxiter=200)
                except RuntimeError:  # pragma: nocover
                    # We get a RuntimeError if the maximum number of
                    # iterations has been reached.
                    raise RuntimeError(
                        "Could not find optimum Lagrange multiplier in 200"
                        " iterations.")

                # xxxxxxxxxx Handle case where a bad mu_i was found xxxxxxx
                # Sometimes the optimization algorithm finds a solution,
                # but it is clearly wrong with a very high value of
                # mu_i. In that case we will do some scaling and try again.
                assert (isinstance(mu_i, float))
                if abs(mu_i) > 1e20:
                    mu_i = optimize.newton(func,
                                           min_mu_i,
                                           args=(sum_term * 10,
                                                 Hii_herm_U * 10, self.P[i]),
                                           maxiter=200)
                    mu_i /= 10.
                    cost = func(mu_i, sum_term, Hii_herm_U, self.P[i])
                    # If our new solution is still bad then we raise a
                    # RuntimeError exception to indicate that a good
                    # solution was not found
                    if cost > self.P[i] / 1e6:
                        # Cost is still positive. The current value for mu
                        # can't be used, since the power restriction will
                        # not be valid. Note that we allow a positive cost
                        # lower then self.P[i]/1e6 since this will be
                        # relatively close to zero.
                        msg = "Could not find a good Lagrange multiplier"
                        raise RuntimeError(msg)
                # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx

                # Now that we have the best value for mu_i, lets calculate Vi
                Vi = self._calc_Vi_for_a_given_mu(sum_term, mu_i, Hii_herm_U)
                # Vi = self._calc_Vi_for_a_given_mu2(
                #     inv_sum_term, mu_i, Hii_herm_U)

                # If any load_factor was added (in case the original
                # sum_term is a singular matrix) we will add it to the
                # optimum mu_i, since this is the effective value of mu_i.
                self._mu[i] = mu_i + load_factor

        # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
        # xxxxxxxxxx Case when the mu value is provided xxxxxxxxxxxxxxxxxxx
        # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
        else:
            self._mu[i] = mu_i
            Vi = self._calc_Vi_for_a_given_mu(sum_term, mu_i, Hii_herm_U)
            # Vi = self._calc_Vi_for_a_given_mu2(inv_sum_term,
            #                                    mu_i, Hii_herm_U)

        return Vi

    def _updateF(self) -> None:
        """
        Updates the precoder of all users.
        """
        # Note that _mu should never be negative. By setting the values to
        # -1 here if after the solve method some element in _mu is still
        # negative then something was not right.
        self._mu = -1.0 * np.ones(self.K, dtype=float)

        Vi = np.zeros(self.K, dtype=np.ndarray)
        norm_Vi = np.zeros(self.K, dtype=np.ndarray)
        for k in range(self.K):
            # Note: The square of the Frobenius norm of Vi is NOT equal to
            # one. Since the full_F property will apply the correct power
            # scaling, the norm of self._F must be equal to one.
            Vi[k] = self._calc_Vi(k)
            norm_Vi[k] = Vi[k] / np.linalg.norm(Vi[k], 'fro')

        self._clear_precoder_filter()
        self._full_F = Vi
        self._F = norm_Vi


# xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
# xxxxxxxxxxxxxxx GreedStreamIASolver Class xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
# xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
class GreedStreamIASolver:
    """
    Implements the Greed Stream Interference Alignment algorithm variation.

    This is not a new IA algorithm, but rather a variation of existing IA
    algorithms. The idea is to use another IA algorithm to find the IA
    solution for the desired maximum number of streams. After the solution
    is found, we remove the worst stream and use the same algorithm again
    to find a new solution. If the solution after the stream reduction
    provided a larger sum capacity we remove the worst stream again and
    keep going until each user has only one stream or the sum capacity
    after stream reduction is lower. The final solution will be for the
    number of streams that yielded the largest sum capacity.

    Parameters
    ----------
    iasolver_obj : T <= IASolverBaseClass
        Must be an object of a derived class of IterativeIASolverBaseClass.
    """
    def __init__(self, iasolver_obj: Type[IASolverBaseClass]):
        self._iasolver = iasolver_obj
        self._runned_iterations = 0

        # #IASolverBaseClass.__init__(self, multiUserChannel)

        # # Maximum number of iterations that the underlying iterative IA
        # # algorithm can run for a given stream configuration.
        # self.max_iterations_per_run = 50

        # # Relative change of the precoder of the underlying algorithm to
        # # consider if the algorithm converged.
        # self.relative_factor = 1e-6

        # Store the full_F, W_H and Ns for the previous stream
        # configuration
        self._old_F: Optional[Sequence[np.ndarray]] = None
        self._old_full_F: Optional[Sequence[np.ndarray]] = None
        self._old_W_H: Optional[Sequence[np.ndarray]] = None
        self._old_Ns: Optional[Sequence[np.ndarray]] = None

    @property
    def runned_iterations(self) -> int:
        """
        Get method for the runned_iterations property.

        Returns
        -------
        int
            The number of runned iterations.
        """
        return self._runned_iterations

    def solve(self,
              Ns: IntOrIntSequence,
              P: Optional[FloatOrFloatSequence] = None) -> int:
        """
        Find the IA solution.

        This method updates the 'F' and 'W' member variables.

        Parameters
        ----------
        Ns : int | np.ndarray
            Number of streams of each user.
        P : np.ndarray | List[float] | float, optional
            Power of each user. If not provided, a value of 1 will be used
            for each user.

        Returns
        -------
        int
            Number of iterations the iterative interference alignment
            algorithm run.
        """
        assert (isinstance(self._iasolver, IterativeIASolverBaseClass))
        self._iasolver.clear()
        self._runned_iterations = 0

        # Find the solution for the number of asked streams. Note that
        # depending of the underlying IA algorithm the number of streams in
        # the solution for some user(s) can be lower then the values in Ns
        self._runned_iterations += self._iasolver.solve(Ns, P)

        # We check if any user has more then one stream, since otherwise we
        # can't remove any stream.
        #
        # The keep_going variable will then indicate if stream reduction should
        # be tried. If there is no user with more then one stream then we have
        # no way to reduce a stream.
        keep_going = bool(np.any(np.greater(self._iasolver.Ns, 1)))

        while keep_going is True:
            # xxxxxxxxxx Store the current solution xxxxxxxxxxxxxxxxxxxxxxx
            self._old_F = [F.copy() for F in self._iasolver.F]
            self._old_full_F = [
                full_F.copy() for full_F in self._iasolver.full_F
            ]
            self._old_W_H = [W_H.copy() for W_H in self._iasolver.W_H]
            self._old_Ns = copy(self._iasolver.Ns)

            old_sum_capacity = self._iasolver.calc_sum_capacity()
            # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx

            # xxxxxxxxxx Find the index of the stream to be removed xxxxxxx
            # Note that you need to have called the solve method of the
            # self._iasolver object before you call this method here.
            user_idx, stream_idx = \
                self._find_index_stream_with_worst_sinr()
            # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx

            # xxxxx Remove the stream and find a new IA solution xxxxxxxxxx
            self._iasolver.F[user_idx] = np.delete(self._iasolver.F[user_idx],
                                                   stream_idx, 1)
            self._iasolver.full_F[user_idx] = np.delete(
                self._iasolver.full_F[user_idx], stream_idx, 1)
            self._iasolver.Ns[user_idx] -= 1

            # Note that the F member variable is the normalized
            # precoder. Since we removed one column, let's normalize it
            # again.
            self._iasolver.F[user_idx] /= np.linalg.norm(
                self._iasolver.F[user_idx], 'fro')

            #
            self._iasolver.initialize_with = 'fix'
            self._runned_iterations += self._iasolver.solve(
                self._iasolver.Ns, self._iasolver.P)
            # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx

            # xxxxxxxxxx Check if the new solution is better xxxxxxxxxxxxxx
            new_sum_capacity = self._iasolver.calc_sum_capacity()

            # If the new solution is not better, we restore the previous
            # solution and set keep_going to False to stop the stream
            # reduction.
            if old_sum_capacity > new_sum_capacity:
                # Lets restore the previous solution. First we clear the
                # current solution.
                self._iasolver.clear()

                # Now we set the precoders
                self._iasolver.set_precoders(F=self._old_F,
                                             full_F=self._old_full_F,
                                             P=P)
                self._iasolver.set_receive_filters(W_H=self._old_W_H)

                keep_going = False
            # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx

            # xxxxx Check if at least one user has more then 1 stream xxxxx
            # noinspection PyTypeChecker
            if not np.any(np.greater(self._iasolver.Ns, 1)):
                # If there is no user with more then 1 stream, then we
                # can't reduce streams anymore. Let's stop the while loop
                # then.
                keep_going = False  # pragma: no cover
            # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx

        return self.runned_iterations

    def _find_index_stream_with_worst_sinr(self) -> Tuple[int, int]:
        """
        Considering the current solution (precoders and receive filters) in
        self._iasolver, find the index of the user and stream corresponding
        to the worst SINR.

        Returns
        -------
        user_idx : int
            The index of the user that has the stream with the worst SINR.
        stream_idx : int
            The index of the stream (for user `user_idx`) with the worst
            SINR.
        """
        sinrs = self._iasolver.calc_SINR()  # type: ignore
        # First we find the index of the minimum SINR for each user
        min_sinr_indexes = [np.argmin(s) for s in sinrs]

        # For each user we get his minimum sinr.
        min_sinrs = [
            sinrs[i][min_sinr_indexes[i]]
            for i in range(self._iasolver.K)  # type: ignore
        ]

        # Index of the users in ascending order of the sinrs. The fist
        # element is the index of the user with the minimum SINR.
        min_sinr_user_idx = np.argsort(min_sinrs)

        # Let's discard any user that has only one stream, since we can't
        # reduce the number of streams of that user.
        valid_users_idx = \
            np.arange(self._iasolver.K)[self._iasolver.Ns > 1]  # type: ignore
        min_sinr_user_idx = [
            i for i in min_sinr_user_idx if i in valid_users_idx
        ]

        user_idx = min_sinr_user_idx[0]
        stream_idx = min_sinr_indexes[user_idx]
        return user_idx, stream_idx


# xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
# xxxxxxxxxxxxxxx BruteForceStreamIASolver xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
# xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
class BruteForceStreamIASolver:
    """
    Implements the Brute Force Stream Interference Alignment algorithm
    variation.

    This is not a new IA algorithm, but rather a variation of existing IA
    algorithms. The idea is to use another IA algorithm to find the IA
    solution for each possible stream configuration (number of streams for
    each user) and keep the best one.

    Note: IA algorithms can provide different performance for the same
    number of streams but with different initializations. To reduce this
    variability we set the initialization method to 'svd' (see
    'initialize_with' property in the IterativeIASolverBaseClass class) so
    that the initialization is always the same.

    Parameters
    ----------
    iasolver_obj : T <= IASolverBaseClass
        Must be an object of a derived class of IterativeIASolverBaseClass.
    """
    def __init__(self, iasolver_obj: IterativeIASolverBaseClass):
        self._iasolver = iasolver_obj
        self._runned_iterations = 0

        # store every possible stream combination
        self._stream_combinations: List[int] = []
        # store sum capacity for each stream combination
        self._every_sum_capacity: List[float] = []

        # Store the full_F, W_H and Ns for the previous stream
        # configuration
        self._best_F: Optional[Sequence[np.ndarray]] = None
        self._best_full_F: Optional[Sequence[np.ndarray]] = None
        self._best_W_H: Optional[Sequence[np.ndarray]] = None
        self._best_Ns: Optional[Sequence[np.ndarray]] = None

    def clear(self) -> None:
        """
        Clear the BruteForceStreamIASolver object.
        """
        # store every possible stream combination
        self._stream_combinations = []
        # store sum capacity for each stream combination
        self._every_sum_capacity = []

        # Store the full_F, W_H and Ns for the previous stream
        # configuration
        self._best_F = None
        self._best_full_F = None
        self._best_W_H = None
        self._best_Ns = None

    @property
    def runned_iterations(self) -> int:
        """
        Get method for the runned_iterations property.

        Returns
        -------
        int
            The number of runned iterations.
        """
        return self._runned_iterations

    @property
    def stream_combinations(self) -> Sequence[int]:
        """
        Get method for the stream_combinations property.

        Returns
        -------
        tuple
            Tuple containing every possible stream combination.
        """
        return self._stream_combinations

    @property
    def every_sum_capacity(self) -> List[float]:
        """
        Get method for the every_sum_capacity property.

        Returns
        -------
        list
            Tuple containing the sum capacity for each stream
            combination in `self.stream_combinations`.
        """
        return self._every_sum_capacity

    def solve(self,
              Ns: IntOrIntSequence,
              P: Optional[FloatOrFloatSequence] = None) -> int:
        """
        Find the IA solution.

        This method updates the 'F' and 'W' member variables.

        Parameters
        ----------
        Ns : int | np.ndarray
            MAXIMUM number of streams of each user. All possible values
            from 1 to Ns (or Ns[k], if Ns is an array) will tried.
        P : np.ndarray | List[float] | float, optional
            Power of each user. If not provided, a value of 1 will be used
            for each user.

        Returns
        -------
        int
            Number of iterations the iterative interference alignment
            algorithm run.
        """
        self._iasolver.clear()
        self._runned_iterations = 0

        self._iasolver.initialize_with = 'svd'
        K = self._iasolver.K

        if isinstance(Ns, int):
            Ns = np.ones(K, dtype=int) * Ns
        assert (not isinstance(Ns, int))

        # xxxxxxxxxx Find all possible stream configurations xxxxxxxxxxxxxx
        # First we create a list of K lists, where each inner list has the
        # possible number of streams of one user.
        # Ex: If Ns is [2, 2, 3] then the list of lists will be
        # [[1, 2], [1, 2], [1, 2, 3]]
        each_user_variation = [range(1, Ns[i] + 1) for i in range(K)]

        # Calculate all possible combinations of the inner lists.
        self._stream_combinations = list(
            product(*each_user_variation))  # type: ignore

        # xxxxx Find the solution for each stream configuration xxxxxxxxxxx
        self._every_sum_capacity = []

        # xxxxx Compute the solution for the first combination xxxxxxxxxxxx
        stream_combinations = iter(self._stream_combinations)
        comb = next(stream_combinations)
        self._iasolver.clear()
        self._runned_iterations += self._iasolver.solve(np.array(comb), P)
        self._every_sum_capacity.append(self._iasolver.calc_sum_capacity())

        self._best_F = self._iasolver._F
        self._best_full_F = self._iasolver._full_F
        self._best_W_H = self._iasolver._W_H
        self._best_Ns = self._iasolver.Ns
        best_sum_capacity = self._every_sum_capacity[-1]
        # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx

        # xxxxx Compute the solution for the remaining combinations xxxxxxx
        for comb in stream_combinations:
            self._iasolver.clear()
            self._runned_iterations += \
                self._iasolver.solve(np.array(comb), P)
            self._every_sum_capacity.append(self._iasolver.calc_sum_capacity())

            # If the current solution is better then the best one, store it
            # as the new best solution.
            if self._every_sum_capacity[-1] > best_sum_capacity:
                best_sum_capacity = self._every_sum_capacity[-1]
                self._best_F = self._iasolver._F
                self._best_full_F = self._iasolver._full_F
                self._best_W_H = self._iasolver._W_H
                self._best_Ns = self._iasolver.Ns
        # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx

        # Now that we tested every possible solution, lets keep the best
        # one we found
        self._iasolver.clear()
        self._iasolver._F = self._best_F
        self._iasolver._full_F = self._best_full_F
        self._iasolver._W_H = self._best_W_H
        self._iasolver._Ns = self._best_Ns
        # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx

        return self._runned_iterations


# xxxxxxxxxx End of the File xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pyphysim',
 'pyphysim.c_extensions',
 'pyphysim.cell',
 'pyphysim.channel_estimation',
 'pyphysim.channels',
 'pyphysim.comm',
 'pyphysim.extra',
 'pyphysim.extra.MATLAB',
 'pyphysim.ia',
 'pyphysim.mimo',
 'pyphysim.modulators',
 'pyphysim.reference_signals',
 'pyphysim.simulations',
 'pyphysim.subspace',
 'pyphysim.util']

package_data = \
{'': ['*']}

install_requires = \
['configobj', 'matplotlib', 'numba>=0.48.0,<0.49.0', 'numpy', 'scipy']

setup_kwargs = {
    'name': 'pyphysim',
    'version': '0.4',
    'description': 'Implementation of a digital communication (physical layer) in python',
    'long_description': '![Testing](https://travis-ci.org/darcamo/pyphysim.svg?branch=master)\n[![Coverage Status](https://coveralls.io/repos/github/darcamo/pyphysim/badge.svg?branch=master)](https://coveralls.io/github/darcamo/pyphysim?branch=master)\n[![Documentation Status](https://readthedocs.org/projects/pyphysim/badge/?version=latest)](http://pyphysim.readthedocs.io/en/latest/?badge=latest)\n[![pre-commit](https://img.shields.io/badge/pre--commit-enabled-brightgreen?logo=pre-commit&logoColor=white)](https://github.com/pre-commit/pre-commit)\n\nPyPhysim\n========\n\nSimulation of Digital Communication (physical layer) in Python.\n\nThis includes classes related to digital modulation, AWGN channels, MIMO,\nOFDM, etc.. It also includes classes related to multiuser transmission such\nas block diagonalization, interference alignment, etc.\n\nFurthermore, a framework for implementing Monte Carlo simulations is also\nimplemented (see the pyphysim.simulations package).\n\n\nNote\n----\n\nInstall [poetry](https://python-poetry.org/), clone this repository and then use\nthe command `poetry install` to install pyphysim.\n\nYou can also directly install it from pypi with `pip install pyphysim`.\n',
    'author': 'Darlan Cavalcante Moreira',
    'author_email': 'darcamo@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/darcamo/pyphysim',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



#!/usr/bin/env python

# pylint: disable=E1101
"""
Implement classes to represent the progress of a task.

Use the ProgressbarText class for tasks that do not use multiprocessing,
and the ProgressbarMultiProcessServer class for tasks using
multiprocessing.

Basically, the task code must call the "progress" function to update the
progress bar and pass a number equivalent to the increment in the progress
since the last call. The progressbar must know the maximum value equivalent
to all the progress, which is passed during object creator for
ProgressbarText class.

The ProgressbarMultiProcessServer is similar to ProgressbarText class,
accounts for the progress of multiple processes. For each process you need
to call the register_client_and_get_proxy_progressbar to get a proxy
progressbar, where the maximum value equivalent to all the progress that
will be performed by that process is passed in this proxy creation. Each
process then calls the progress method of the proxy progressbar.

Note that there is also a DummyProgressbar whose progress function does
nothing. This is useful when you want to give the user a choice to show or
not the progressbar such that the task code can always call the progress
method and you only change the progressbar object.
"""

import multiprocessing
import os
import sys
import time
import warnings
from typing import Any, List, Optional, Tuple, cast

from IPython.display import display
from ipywidgets import FloatProgress, HBox, Label

from ..util.misc import pretty_time

try:
    # noinspection PyUnresolvedReferences
    import zmq
except ImportError:  # pragma: no cover
    # We don't have a fallback for zmq, but only the
    # ProgressbarZMQServer and ProgressbarZMQClient classes require it
    pass

# Type used to store IP address and port number
ClientID = int
IPAddress = str
PortNumber = int

__all__ = [
    'DummyProgressbar', 'ProgressBarBase', 'ProgressbarText',
    'ProgressbarText2', 'ProgressbarText3', 'ProgressbarMultiProcessServer',
    'ProgressbarMultiProcessClient', 'ProgressbarZMQServer',
    'ProgressBarIPython', 'center_message', 'ProgressbarTextBase',
    'ProgressbarDistributedServerBase', 'ProgressbarDistributedClientBase'
]


# If this function is ever used outside this file, then move it to the
# util.misc module.
def center_message(message: str,
                   length: int = 50,
                   fill_char: str = ' ',
                   left: str = '',
                   right: str = '') -> str:
    """
    Return a string with `message` centralized and surrounded by
    `fill_char`.

    Parameters
    ----------
    message: str
        The message to be centered.
    length : int
        Total length of the centered message (original + any fill).
    fill_char : str
        Filling character.
    left : str
        Left part of the filling.
    right : str
       Right part of the filling.

    Returns
    -------
    cent_message : str
        The centralized message.

    Examples
    --------
    >>> print(center_message("Hello World", 50, '-', 'Left', 'Right'))
    Left-------------- Hello World --------------Right
    """
    message_size = len(message)
    left_size = len(left)
    right_size = len(right)
    fill_size = (length - (message_size + 2) - left_size - right_size)
    left_fill_size = fill_size // 2 + (fill_size % 2)
    right_fill_size = (fill_size // 2)

    new_message = u"{0}{1} {2} {3}{4}".format(left, fill_char * left_fill_size,
                                              message,
                                              fill_char * right_fill_size,
                                              right)
    return new_message


# xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
# xxxxxxxxxxxxxxx DummyProgressbar - START xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
# xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
# noinspection PyMethodMayBeStatic
class DummyProgressbar:  # pragma: no cover
    """
    Dummy progress bar that don't really do anything.

    The idea is that it can be used in place of the :class:`ProgressbarText`
    class, but without actually doing anything.

    See also
    --------
    ProgressbarText
    """
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """
        Initializes the DummyProgressbar object.

        This method accepts any argument without errors, but they won't
        matter, since this class does nothing.
        """

    def progress(self, count: int) -> None:
        """This `progress` method has the same signature from the one in the
        :class:`ProgressbarText` class.

        Nothing happens when this method is called.

        Parameters
        ----------
        count : int
            Ignored
        """


# xxxxxxxxxx DummyProgressbar - END xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx


class ProgressBarBase:
    """
    Base class for all ProgressBar classes.

    Parameters
    ----------
    finalcount : int
        The total amount that corresponds to 100%. Each time the progress
        method is called with a number that number is added with the
        current amount in the progressbar. When the amount becomes equal to
        `finalcount` the bar will be 100% complete.

    Notes
    -----
    Derived classes should implement :func:`_update_iteration` and
    :func:`_display_current_progress`. Optionally derived class might also
    implement :func:`_perform_initialization` and :func:`_perform_finalizations`
    """
    def __init__(self, finalcount: int):
        """
        Initializes the progressbar object.

        Parameters
        ----------
        finalcount : int
            The total amount that corresponds to 100%. Each time the
            progress method is called with a number that number is added
            with the current amount in the progressbar. When the amount
            becomes equal to `finalcount` the bar will be 100% complete.
        """
        self.finalcount = finalcount

        # This will be set to True after the `start` method is called to
        # initialize the progressbar.
        self._initialized = False

        # This will be set to true when the progress reaches 100%. When
        # this is True, any subsequent calls to the `progress` method will
        # be ignored.
        self._finalized = False

        # This variable will store the time when the `start` method was
        # called for the first time (either manually or in the `progress`
        # method. It will be used for tracking the elapsed time.
        self._start_time = 0.0
        # This variable will store the time when the `stop` method was
        # called for the first time (either manually or in the `progress`
        # method. It will be used for tracking the elapsed time.
        self._stop_time = 0.0

    @property
    def elapsed_time(self) -> str:
        """
        Get method for the elapsed_time property.

        Returns
        -------
        str
            The elapsed time.
        """
        elapsed_time = 0.0
        if self._initialized is True:
            if self._finalized is False:
                elapsed_time = time.time() - self._start_time
            else:
                elapsed_time = self._stop_time - self._start_time
        return pretty_time(elapsed_time)

    def _count_to_percent(self, count: int) -> float:
        """
        Convert a given count into the equivalent percentage.

        Parameters
        ----------
        count : int
            The current count to be represented in the progressbar. The
            progressbar represents this count as a percent value of
            self.finalcount

        Returns
        -------
        percentage : float
            The percentage that `count` is of self.finalcount (between 0
            and 100)
        """
        percentage = (count / float(self.finalcount)) * 100.0
        return percentage

    def _perform_initialization(self) -> None:
        """
        Perform any initializations for the progressbar.

        This method should be implemented in sub-classes if any
        initialization code should be run.
        """

    def _perform_finalizations(self) -> None:  # pragma: nocover
        """
        Perform any finalization (cleanings) after the progressbar stops.

        This method should be implemented in sub-classes if any
        finalization code should be run.
        """

    def start(self) -> None:
        """
        Start the progressbar.

        This method should be called just before the progressbar is used
        for the first time. Among possible other things, it will store the
        current time so that the elapsed time can be tracked.
        """
        if self._initialized is False:
            self._start_time = time.time()
            self._perform_initialization()
            self._initialized = True

    def stop(self) -> None:
        """
        Stop the progressbar.

        This method is automatically called in the `progress` method when
        the progress reaches 100%. If manually called, any subsequent call
        to the `progress` method will be ignored.
        """
        if self._finalized is False:
            self._stop_time = time.time()

            # When progress reaches 100% we set the internal variable
            # to True so that any subsequent calls to the `progress`
            # method will be ignored.
            self._finalized = True

            self._perform_finalizations()

    # pylint:disable=R0201,W0613
    def _update_iteration(self, count: int) -> None:  # pragma: no cover
        """
        Update the progressbar according with the new `count`.

        Parameters
        ----------
        count : int
            The current count to be represented in the progressbar. The
            progressbar represents this count as a percent value of
            self.finalcount
        """
        raise NotImplementedError("Implement this method in a subclass")

    def _display_current_progress(self) -> None:  # pragma: nocover
        """
        Refresh the progress representation.

        This method should be defined in a subclass.
        """
        raise NotImplementedError("Implement this method in a subclass")

    def progress(self, count: int) -> None:
        """
        Updates the current progress.

        Calling this function will update the the current progress.

        Parameters
        ----------
        count : int
            The current count to be represented in the progressbar. The
            progressbar represents this count as a percent value of
            self.finalcount

        Notes
        -----
        How the progressbar is actually represented depends on the
        subclass.  In the subclasses implement the `_update_iteration`
        method to update the current representation of the progressbar and
        the `_update_progress_display` to actually display the current
        progress.
        """
        if self._finalized is False:
            # Start the progressbar. This only have an effect the first
            # time it is called. It initializes the elapsed time tracking
            # and call the _perform_initialization method to perform any
            # initialization.
            self.start()

            # Sanity check. If count is greater then self.finalcount we set
            # it to self.finalcount
            if count > self.finalcount:
                count = self.finalcount

            # Update the progressbar representation. this is up to the
            # subclass. Note that this method should not refresh the
            # progressbar, which is left to the _display_current_progress
            # method.
            self._update_iteration(count)

            # Refresh the progress representation.
            self._display_current_progress()

            # If count is equal to self.finalcount we have reached
            # 100%. In that case, we also write a final newline
            # character.
            if count == self.finalcount:
                self.stop()


# xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
# xxxxxxxxxxxxxxx ProgressbarTextBase - START xxxxxxxxxxxxxxxxxxxxxxxxxxxxx
# xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
# The code here and in some of the derived classes is inspired in the code
# located in
# http://nbviewer.ipython.org/url/github.com/ipython/ipython/raw/master/
# /examples/notebooks/Progress%20Bars.ipynb
#
# noinspection PyAbstractClass
class ProgressbarTextBase(ProgressBarBase):  # pylint: disable=R0902,W0223
    """
    Base class for Text progressbars.

    Parameters
    ----------
    finalcount : int
        The total amount that corresponds to 100%. Each time the progress
        method is called with a number that number is added with the
        current amount in the progressbar. When the amount becomes equal to
        `finalcount` the bar will be 100% complete.
    progresschar : str, optional
        The character used to represent progress.
    message : str, optional
        A message to be shown in the top of the progressbar.
    output : File like object
        Object with a 'write' method, which controls where the progress-bar
        will be printed. By default sys.stdout is used, which means that
        the progress will be printed in the standard output.

    Notes
    -----
    Derived classes must implement at least `_update_iteration` and this
    method should update the `prog_bar` member variable with the text
    representation of the progress.
    """
    def __init__(self,
                 finalcount: int,
                 progresschar: str = '*',
                 message: str = '',
                 output: Any = sys.stdout) -> None:
        """
        Initializes the progressbar object.

        Parameters
        ----------
        finalcount : int
            The total amount that corresponds to 100%. Each time the
            progress method is called with a number that number is added
            with the current amount in the progressbar. When the amount
            becomes equal to `finalcount` the bar will be 100% complete.
        progresschar : str, optional (default to '*')
            The character used to represent progress.
        message : str, optional
            A message to be shown in the top of the progressbar.
        output : File like object
            Object with a 'write' method, which controls where the
            progress-bar will be printed. By default sys.stdout is used,
            which means that the progress will be printed in the standard
            output.
        """
        super(ProgressbarTextBase, self).__init__(finalcount)

        # This will be updated with the progress and should contain the
        # whole string representation of the progressbar.
        self.prog_bar = "Progress Representation"
        self.old_prog_bar = ""

        # character used to represent progress
        self.progresschar = progresschar

        # If output points to a file (and not to stdout) and this is set to
        # True, then the file will be erased after the progress finishes.
        self.delete_progress_file_after_completion = False

        # This should be a multiple of 10. The lower possible value is 40.
        self._width = 50

        # By default, self._output points to sys.stdout so I can use the
        # write/flush methods to display the progress bar.
        self._output = output
        self._message = message  # THIS WILL BE IGNORED

        # If true, an empty line will be printed when the progress finishes
        self._print_empty_line_at_the_end = True

    def __del__(self) -> None:
        """
        Delete the output file if there is any and
        delete_progress_file_after_completion was set to True when the
        progressbar object is deleted.
        """
        # In case the progressbar object is deleted before the progress
        # finishes the `stop` method will not be called and thus the output
        # file (if there is any) would not be deleted even if
        # delete_progress_file_after_completion was set to True. Therefore,
        # we implement the __del__ method here to call the
        # _maybe_delete_output_file method to do that.
        self._maybe_delete_output_file()

    # xxxxxxxxxx width property xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
    # Text progressbars have a width property indicating the width (in
    # number of characters) of the full progress.
    @property
    def width(self) -> int:
        """Get method for the width property."""
        return self._width

    @width.setter
    def width(self, value: int) -> None:
        """
        Set method for the width property.

        Parameters
        ----------
        value : int
        """
        # If value is lower than 40, the width will be set to 40.
        # If value is not a multiple of 10, width will be set to the
        # largest multiple of 10 which is lower then value.
        if value < 40:
            value = 40
        self._width = value - (value % 10)

    # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx

    def _get_percentage_representation(self,
                                       percent: float,
                                       central_message: str = '{percent}%',
                                       left_side: str = '[',
                                       right_side: str = ']') -> str:
        """
        Get the percent representation as a string suitable to the text
        progressbar.

        Parameters
        ----------
        percent : float
            The percentage to be represented.
        central_message : str
            A message that will be in the middle of the percentage bar. If
            there is the label '{percent}' in the central_message it will
            be replaced by the percentage. If there is the label
            '{elapsed_time}' in the central_message it will be replaced by
            the elapsed time. Note that this message should be very small,
            since it hides the progresschars.
        left_side : str
            The left side of the bar.
        right_side : str
            The right side of the bar.

        Returns
        -------
        representation : str
            A string with the representation of the percentage.
        """
        # Remove any fractional part
        percent_done = int(percent)
        elapsed_time = self.elapsed_time

        # Calculates how many characters are spent just for the sides.
        sides_length = len(left_side) + len(right_side)
        # The width should be large enough to contain both the left_side
        # and right_side and still have (reasonable) enough space for the
        # characters representing the progress.
        assert self.width > sides_length + 20

        # Space that will be used bu the characters representing the
        # progress
        all_full = self.width - sides_length

        # Calculates how many characters will be used to represent the
        # `percent_done` value
        num_hashes = int((percent_done / 100.0) * all_full)

        prog_bar = (left_side + self.progresschar * num_hashes + ' ' *
                    (all_full - num_hashes) + right_side)

        # Replace the center of prog_bar with the message
        central_message = central_message.format(percent=percent_done,
                                                 elapsed_time=elapsed_time)
        pct_place = (len(prog_bar) // 2) - (len(str(central_message)) // 2)
        prog_bar = prog_bar[0:pct_place] + central_message + prog_bar[
            pct_place + len(central_message):]

        return prog_bar

    def _maybe_delete_output_file(self) -> None:
        """
        Delete the output file (if there is any) when
        delete_progress_file_after_completion is set to True.
        """
        if self.delete_progress_file_after_completion is True:
            # Try to get the file name associated with the output. If we
            # can get an actual file and
            # delete_progress_file_after_completion is set to True we will
            # delete that file
            try:
                name = self._output.name
                # We will only delete a file if the name does not point to
                # stdout.
                if name != '<stdout>':
                    try:
                        os.remove(name)
                    except OSError:  # Pragma: no cover
                        pass

            except AttributeError:  # pragma: no cover
                # If an attribute error was raised then the output is not a
                # file like object and therefore we don't need to delete
                # any file
                pass

    def _perform_finalizations(self) -> None:
        """
        Perform any finalization (cleanings) after the progressbar stops.
        """
        if self._print_empty_line_at_the_end is True:
            # Print an empty line after the last iteration to be consistent
            # with the ProgressbarText class
            self._output.write(u"\n")

            # Flush everything to guarantee that at this point everything
            # is written to the output.
            self._output.flush()

            # This will only delete the output file if self._output
            # actually points to a file and if
            # self.delete_progress_file_after_completion is set to True
            self._maybe_delete_output_file()

    def _display_current_progress(self) -> None:
        """
        Refresh the progress representation.

        All text progressbars should implement the `_update_iteration` to
        update the `prog_bar` member variable with the text representation
        of the progressbar.

        This method is responsible to sending this text representation to
        the output.
        """
        # We will only write the progress if it actually changed since
        # the last time. This is specially useful when the output is a
        # file and it will avoid writing many unnecessary equal lines to
        # the file.
        if self.old_prog_bar != self.prog_bar:
            # Save the current prog_bar variable before it is updated in
            # the _update_iteration method.
            self.old_prog_bar = self.prog_bar

            # We simple change the cursor to the beginning of the line and
            # write the string representation of the prog_bar variable.
            self._output.write(u'\r')
            self._output.write(u'{0}'.format(self.prog_bar))

            # Flush everything to guarantee that at this point
            # everything is written to the output.
            self._output.flush()

    # The string representation of a text progressbar should display the
    # whole progressbar
    def __str__(self) -> str:
        return str(self.prog_bar)


# xxxxxxxxxx ProgressbarTextBase - END xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx


# xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
# xxxxxxxxxxxxxxx ProgressbarText - START xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
# xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
# http://
# code.activestate.com/recipes/299207-console-text-progress-indicator-class/
#
# CLASS NAME: ProgressbarText
#
# Original Author of the ProgressbarText class:
# Larry Bates (lbates@syscononline.com)
# Written: 12/09/2002
#
# Modified by Darlan Cavalcante Moreira in 10/18/2011
# Released under: GNU GENERAL PUBLIC LICENSE
class ProgressbarText(ProgressbarTextBase):
    """
    Class that prints a representation of the current progress as text.

    You can set the final count for the progressbar, the character that
    will be printed to represent progress and a small message indicating
    what the progress is related to.

    In order to use this class, create an object outsize a loop and inside
    the loop call the `progress` function with the number corresponding to
    the progress (between 0 and finalcount). Each time the `progress`
    function is called a number of characters will be printed to show the
    progress. Note that the number of printed characters correspond is
    equivalent to the progress minus what was already printed.

    See also
    --------
    DummyProgressbar

    Examples
    --------
    >> pb = ProgressbarText(100, 'o', "Hello Simulation")
    >> pb.start()
    ---------------- Hello Simulation ---------------1
        1    2    3    4    5    6    7    8    9    0
    ----0----0----0----0----0----0----0----0----0----0
    >> pb.progress(20)
    oooooooooo
    >> pb.progress(40)
    oooooooooooooooooooo
    >> pb.progress(50)
    ooooooooooooooooooooooooo
    >> pb.progress(100)
    oooooooooooooooooooooooooooooooooooooooooooooooooo
    """
    def __init__(self,
                 finalcount: int,
                 progresschar: str = '*',
                 message: str = '',
                 output: Any = sys.stdout):
        """
        Initializes the ProgressbarText object.

        Parameters
        ----------
        finalcount : int
            The total amount that corresponds to 100%. Each time the
            progress method is called with a number that number is added
            with the current amount in the progressbar. When the amount
            becomes equal to `finalcount` the bar will be 100% complete.
        progresschar : str, optional (default to '*')
            The character used to represent progress.
        message : str, optional
            A message to be shown in the top of the progressbar.
        output : File like object
            Object with a 'write' method, which controls where the
            progress-bar will be printed. By default sys.stdout is used,
            which means that the progress will be printed in the standard
            output.
        """
        ProgressbarTextBase.__init__(self, finalcount, progresschar, message,
                                     output)

        # stores how many characters where already printed in a previous
        # call to the `progress` function
        self.progresscharcount = 0

    def __get_initialization_bartitle(self) -> str:
        """
        Get the progressbar title.

        The title is the first line of the progressbar initialization
        message.

        The bar title is something like the line below

        ------------------- % Progress ------------------1\n

        when there is no message.

        Returns
        -------
        str
            The bar title.

        Notes
        -----
        This method is only a helper method called in the
        `_perform_initialization` method.
        """
        if len(self._message) != 0:
            message = self._message
        else:
            message = '% Progress'

        bartitle = center_message(message, self.width + 1, '-', '', '1\n')
        return bartitle

    def __get_initialization_markers(self) -> Tuple[str, str]:
        """
        The initialization markers 'mark' the current progress in the
        progressbar that will appear below it.

        Returns
        -------
        Tuple[str, str]
            A tuple containing the 'two lines' with the progress
            markers. That is, (marker_line1, marker_line2)

        Notes
        -----
        This method is only a helper method called in the
        `_perform_initialization` method.
        """
        steps = self.width // 10  # This division must be exact

        values1 = ['1', '2', '3', '4', '5', '6', '7', '8', '9', '0']
        values2 = ['0', '0', '0', '0', '0', '0', '0', '0', '0', '0']

        common = "{0}{1}\n"

        line1sep = ' ' * (steps - 1)
        line1 = common.format(line1sep, line1sep.join(values1))

        line2sep = '-' * (steps - 1)
        line2 = common.format(line2sep, line2sep.join(values2))

        return line1, line2

    def _perform_initialization(self) -> None:
        bartitle = self.__get_initialization_bartitle()
        marker_line1, marker_line2 = self.__get_initialization_markers()

        self._output.write(bartitle)
        self._output.write(marker_line1)
        self._output.write(marker_line2)

    def _update_iteration(self, count: int) -> None:
        percentage = self._count_to_percent(count)

        # Set the self.prog_bar variable simply as a string containing as
        # many self.progresschar characters as necessary.
        self.prog_bar = self._get_percentage_representation(percentage,
                                                            left_side='',
                                                            right_side='',
                                                            central_message='')


# xxxxxxxxxx ProgressbarText - END xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx


# xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
# xxxxxxxxxxxxxxx ProgressbarText2 - START xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
# xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
class ProgressbarText2(ProgressbarTextBase):
    """
    Class that prints a representation of the current progress as text.

    You can set the final count for the progressbar, the character that
    will be printed to represent progress and a small message indicating
    what the progress is related to.

    In order to use this class, create an object outsize a loop and inside
    the loop call the `progress` function with the number corresponding to
    the progress (between 0 and finalcount). Each time the `progress`
    function is called a number of characters will be printed to show the
    progress. Note that the number of printed characters correspond is
    equivalent to the progress minus what was already printed.

    Parameters
    ----------
    finalcount : int
        The total amount that corresponds to 100%. Each time the progress
        method is called with a number that number is added with the
        current amount in the progressbar. When the amount becomes equal to
        `finalcount` the bar will be 100% complete.
    progresschar : str, optional (default to '*')
        The character used to represent progress.
    message : str, optional
        A message to be shown in the right of the progressbar. If this
        message contains "{elapsed_time}" it will be replaced by the
        elapsed time.
    output : File like object
        Object with a 'write' method, which controls where the progress-bar
        will be printed. By default sys.stdout is used, which means that
        the progress will be printed in the standard output.
    """
    def __init__(self,
                 finalcount: int,
                 progresschar: str = '*',
                 message: str = '',
                 output: Any = sys.stdout) -> None:
        """
        Initializes the progressbar object.

        Parameters
        ----------
        finalcount : int
            The total amount that corresponds to 100%. Each time the
            progress method is called with a number that number is added
            with the current amount in the progressbar. When the amount
            becomes equal to `finalcount` the bar will be 100% complete.
        progresschar : str, optional (default to '*')
            The character used to represent progress.
        message : str, optional
            A message to be shown in the right of the progressbar. If this
            message contains "{elapsed_time}" it will be replaced by the
            elapsed time.
        output : File like object
            Object with a 'write' method, which controls where the
            progress-bar will be printed. By default sys.stdout is used,
            which means that the progress will be printed in the standard
            output.
        """
        ProgressbarTextBase.__init__(self, finalcount, progresschar, message,
                                     output)

    def _update_iteration(self, count: int) -> None:
        """
        Update the self.prog_bar member variable according with the new
        `count`.

        Parameters
        ----------
        count : int
            The current count to be represented in the progressbar. The
            progressbar represents this count as a percent value of
            self.finalcount
        """
        # Convert `count` to the equivalent percentage
        percent_count = self._count_to_percent(count)

        # Update the self.prog_bar variable with (only) the current
        # percentage representation. The message is not included and will
        # be appended after this.
        self.prog_bar = self._get_percentage_representation(
            percent_count,
            central_message='{percent}%',
            left_side='[',
            right_side=']')

        # Append the message to the self.prog_bar variable if there is one
        # (or a default message if there is no message set)..
        if len(self._message) != 0:
            message = self._message.format(elapsed_time=self.elapsed_time)
            self.prog_bar += "  {0}".format(message)
        else:
            self.prog_bar += '  %d of %d complete' % (count, self.finalcount)


# xxxxxxxxxx ProgressbarText2 - END xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx


# xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
# xxxxxxxxxxxxxxx ProgressbarText3 - START xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
# xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
class ProgressbarText3(ProgressbarTextBase):
    """
    Class that prints a representation of the current progress as text.

    You can set the final count for the progressbar, the character that
    will be printed to represent progress and a small message indicating
    what the progress is related to.

    In order to use this class, create an object outsize a loop and inside
    the loop call the `progress` function with the number corresponding to
    the progress (between 0 and finalcount). Each time the `progress`
    function is called a number of characters will be printed to show the
    progress. Note that the number of printed characters correspond is
    equivalent to the progress minus what was already printed.

     Parameters
    ----------
    finalcount : int
        The total amount that corresponds to 100%. Each time the progress
        method is called with a number that number is added with the
        current amount in the progressbar. When the amount becomes equal to
        `finalcount` the bar will be 100% complete.
    progresschar : str, optional (default to '*')
        The character used to represent progress.
    message : str, optional
        A message to be shown in the progressbar.
    output : File like object
        Object with a 'write' method, which controls where the progress-bar
        will be printed. By default sys.stdout is used, which means that
        the progress will be printed in the standard output.
    """
    def __init__(self,
                 finalcount: int,
                 progresschar: str = '*',
                 message: str = '',
                 output: Any = sys.stdout) -> None:
        """
        Initializes the progressbar object.

        Parameters
        ----------
        finalcount : int
            The total amount that corresponds to 100%. Each time the
            progress method is called with a number that number is added
            with the current amount in the progressbar. When the amount
            becomes equal to `finalcount` the bar will be 100% complete.
        progresschar : str, optional (default to '*')
            The character used to represent progress.
        message : str, optional
            A message to be shown in the progressbar.
        output : File like object
            Object with a 'write' method, which controls where the
            progress-bar will be printed. By default sys.stdout is used,
            which means that the progress will be printed in the standard
            output.
        """
        ProgressbarTextBase.__init__(self, finalcount, progresschar, message,
                                     output)

        # The ProgressbarText3 class already prints an empty line after
        # each update. Therefore, there is no need to print an empty line
        # after all the progress has been finished.
        self._print_empty_line_at_the_end = False

    def _update_iteration(self, count: int) -> None:
        """
        Update the self.prog_bar member variable according with the new
        `count`.

        Parameters
        ----------
        count : int
            The current count to be represented in the progressbar. The
            progressbar represents this count as a percent value of
            self.finalcount
        """
        full_count = "{0}/{1}".format(count, self.finalcount)

        if len(self._message) != 0:
            self.prog_bar = center_message("{0} {1}".format(
                self._message, full_count),
                                           length=self.width,
                                           fill_char=self.progresschar)
        else:
            self.prog_bar = center_message(full_count,
                                           length=self.width,
                                           fill_char=self.progresschar)


# xxxxxxxxxx ProgressbarText3 - END xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx


# xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
# xxxxxxxxxxxxxxx ProgressBarIPython xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
# xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
class ProgressBarIPython(ProgressBarBase):  # pragma: no cover
    """
    Progressbar for IPython notebooks.

    The progressbar will be rendered using IPython widgets.
    """
    def __init__(self,
                 finalcount: int,
                 side_message: Optional[str] = None) -> None:
        """
        Initializes the progressbar object.

        Parameters
        ----------
        finalcount : int
            The total amount that corresponds to 100%. Each time the
            progress method is called with a number that number is added
            with the current amount in the progressbar. When the amount
            becomes equal to `finalcount` the bar will be 100% complete.
        side_message : str
            A message to display on the right side of the progressbar. This
            is rendered as Latex and thus can contain math.
        """
        super(ProgressBarIPython, self).__init__(finalcount)

        # IPython already provide us a nice widget to represent
        # progressbars.
        self.prog_bar = FloatProgress()

        # If `side_message` is provided then we will add the message as a
        # LatexWidget with the message as the value.
        self.side_message = Label()
        if side_message is None:
            self.side_message.visible = False
        else:
            self.side_message.value = side_message
            self.side_message.disabled = True

        # In order to put the float progressbar and the message side by
        # side we use a container.
        self.container_widget = HBox()
        self.container_widget.children = [self.prog_bar, self.side_message]

    def _update_iteration(self, count: int) -> None:
        """
        Update the self.prog_bar member variable according with the new
        `count`.

        Parameters
        ----------
        count : int
            The current count to be represented in the progressbar. The
            progressbar represents this count as a percent value of
            self.finalcount
        """
        percentage = self._count_to_percent(count)

        # Update the IPython progressbar widget
        self.prog_bar.value = percentage

    def _display_current_progress(self) -> None:
        """
        Refresh the progress representation.
        """
        # This method is called everytime the `progress` method is
        # called. However, for progressbar using IPython widgets we only
        # need to display the widget once and IPython will take care of
        # re-displaying it whenever the widget changes. Therefore, we don't
        # need to do anything here and we will display the widget in the
        # `_perform_initialization` method instead, since it is called only
        # once.

    def _perform_initialization(self) -> None:
        """
        Perform any initializations for the progressbar.

        This method should be implemented in sub-classes if any
        initialization code should be run.
        """

        # Display the container with the progressbar and the message.
        # If no message was provided the the text widget inside the
        # container will be invisible
        display(self.container_widget)


# xxxxxxxxxx ProgressBarIPython - END xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx


# xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
# xxxxxxxxxxxxxxx ProgressbarServerBase xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
# xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
# pylint:disable=R0902
class ProgressbarDistributedServerBase:
    """
    Base class for progressbars for distributed computations.

    In order to track the progress of distributed computations two classes
    are required, one that acts as a central point and is responsible to
    actually show the progress (the server), and other class that acts as a
    proxy (the client) and is responsible to sending the current progress
    to the server. There will be one object of the "server class" and one
    or more objects of the "client class", each one tracking the progress
    of one of the distributed computations.

    This class is a base class for the "server part", while the
    :class:`ProgressbarDistributedClientBase` class is a base class for the
    "client part".

    For a full implementation, see the :class:`ProgressbarMultiProcessServer`
    and :class:`ProgressbarMultiProcessClient` classes.

    Parameters
    ----------
    progresschar : str
        Character used in the progressbar.
    message : str
        Message writen in the progressbar.
    sleep_time : float
        Time between progressbar updates (in seconds).
    filename : str
        If filename is None (default) then progress will be output to
        sys.stdout. If it is not None then the progress will be output
        to a file with name `filename`. This is usually useful for
        debugging and testing purposes.
    """
    def __init__(self,
                 progresschar: str = '*',
                 message: str = '',
                 sleep_time: float = 1.0,
                 filename: Optional[str] = None):
        """
        Initializes the ProgressbarDistributedServerBase object.

        Parameters
        ----------
        progresschar : str
            Character used in the progressbar.
        message : str
            Message writen in the progressbar.
        sleep_time : float
            Time between progressbar updates (in seconds).
        filename : str
            If filename is None (default) then progress will be output to
            sys.stdout. If it is not None then the progress will be output
            to a file with name `filename`. This is usually useful for
            debugging and testing purposes.
        """
        self._progresschar = progresschar
        self._message = message

        self._sleep_time = sleep_time
        self._last_id = -1

        self._filename = filename

        self._manager = multiprocessing.Manager()
        self._client_data_list: List[Any] = self._manager.list()  # pylint: disable=E1101

        # total_final_count will be updated each time the register_*
        # function is called.
        #
        # Note that we use a Manager.Value object to store the value
        # instead of using a simple integer because we want modifications
        # to this value to be seem by the other updating process even after
        # start_updater has been called if we are still in the
        # 'start_delay' time.
        # pylint: disable=E1101
        self._total_final_count = self._manager.Value('L', 0)

        # self._update_process will store the process responsible to update
        # the progressbar. It will be created in the first time the
        # start_updater method is called.
        self._update_process: Optional[multiprocessing.Process] = None

        # The event will be set when the process updating the progressbar
        # is running and unset (clear) when it is stopped.
        #
        # Starts unset. Is is set in the _update_progress function
        self.running = multiprocessing.Event()

        # Each time the start_updater method is called this variable is
        # increased by one and each time the stop_updater method is called
        # it is decreased by one. A started update process will only be
        # stopped in the stop_updater method if this variables reaches 0.
        # This control is useful so that we can share the same progressbar
        # for multiple SimulationRunner objects.
        self._start_updater_count = 0

        # # Used for time tracking
        # self._tic = multiprocessing.Value('f', 0.0)
        # self._toc = multiprocessing.Value('f', 0.0)

    @property
    def total_final_count(self) -> int:
        """
        Get method for the total_final_count property.

        Returns
        -------
        int
            The final count.
        """
        return self._total_final_count.get()

    def _update_client_data_list(self) -> None:
        """
        This method process the communication between the client and the
        server.

        It should gather the information sent by the clients (proxy
        progressbars) and update the member variable self._client_data_list
        accordingly, which will then be automatically represented in the
        progressbar.
        """
        # Implement this method in a derived class.

    def register_client_and_get_proxy_progressbar(
            self, total_count: int) -> "ProgressbarDistributedClientBase":
        """
        Register a new "client" for the progressbar and returns a new proxy
        progressbar that the client can use to update its progress by
        calling the `progress` method of this proxy progressbar.

        Parameters
        ----------
        total_count : int
            Total count that will be equivalent to 100% for function.

        Returns
        -------
        obj : Object of a class derived from ProgressbarDistributedClientBase
            The proxy progressbar.
        """
        # Implement this method in a derived class
        #
        # Note: The first thing the implementation of this method in a
        # derived class must do is call the _register_client method to
        # register the new client and get its client_id, like the example
        # below
        # >>> client_id = self._register_client(total_count)
        #
        # After that the implementation of
        # register_client_and_get_proxy_progressbar can create the
        # corresponding proxy progressbar passing the client_id and any
        # other required data.

    def _register_client(self, total_count: int) -> ClientID:
        """
        Register a new "client" for the progressbar and return its `client_id`.

        These returned values must be passed to the corresponding proxy
        progressbar.

        Parameters
        ----------
        total_count : int
            Total count that will be equivalent to 100% progress for the
            function.

        Returns
        -------
        client_id : ClientID
            The client_id. The function whose process is tracked by the
            ProgressbarMultiProcessServer must update the element
            `client_id` with the current
            count.
        """
        # Set self._total_final_count to the value currently stored plus
        # total_count.
        # Remember that the self._total_final_count variable is actually a
        # proxy to the true value (that is, self._total_final_count is a
        # multiprocessing.Manager.Value object). That is way we need the
        # two lines below.
        total_final_count = self.total_final_count
        self._total_final_count.set(total_final_count + total_count)

        # Update the last_id
        self._last_id += 1

        # client_id that will be used by the function
        client_id = self._last_id

        self._client_data_list.append(0)
        return client_id

    # This method will be run in a different process. Because of this the
    # coverage program does not see that this method in run in the test code
    # even though we know it is run (otherwise no output would
    # appear). Therefore, we put the "pragma: no cover" line in it
    def _update_progress(self,
                         filename: Optional[str] = None,
                         start_delay: float = 0.0) -> None:  # pragma: no cover
        """
        Collects the progress from each registered proxy progressbar and
        updates the actual visible progressbar.

        Parameters
        ----------
        filename : str
            Name of a file where the data will be written to. If this is
            None then all progress will be printed in the standard output
            (default)
        start_delay : float, optional
            Delay in seconds before starting the progressbar. During this
            time it is still possible to register new clients and the
            progressbar will only be shown after this delay..
        """
        if start_delay > 0.0:
            time.sleep(start_delay)

        if self.total_final_count == 0:
            warnings.warn('No clients registered in the progressbar')

        if filename is None:
            output = sys.stdout
        else:
            output = open(filename, 'w')

        pbar = ProgressbarText2(self.total_final_count,
                                self._progresschar,
                                self._message,
                                output=output)
        count = 0
        while count < self.total_final_count and self.running.is_set():
            time.sleep(self._sleep_time)
            # Gather information from all client proxybars and update the
            # self._client_data_list member variable
            self._update_client_data_list()

            # Calculates the current total count from the
            # self._client_data_list
            count = sum(self._client_data_list)

            # Maybe a new client was registered. In that case
            # self.total_final_count changed and we need to reflect this in
            # the pbar.finalcount variable.
            pbar.finalcount = self.total_final_count

            # Represents the current total count in the progressbars
            pbar.progress(count)

        # If the self.running event was cleared (because the stop_updater
        # method was called) we most likely exited the while loop before
        # the progressbar was full (count is lower then the total final
        # count). If that is the case, let's set the progressbar to full
        # here.
        if count < self.total_final_count:
            pbar.progress(self.total_final_count)

        # It may exit the while loop in two situations: if count reached
        # the maximum allowed value, in which case the progressbar is full,
        # or if the self.running event was cleared in another
        # process. Since in the first case the event is still set, we clear
        # it here to have some consistence (a cleared event will always
        # mean that the progressbar is not running).
        self.running.clear()
        # self._toc.value = time.time()

        if self._filename is not None:
            output.close()

    def start_updater(self, start_delay: float = 0.0) -> None:
        """
        Start the process that updates the progressbar.

        Parameters
        ----------
        start_delay : float
            Delay in seconds before starting the progressbar. During this
            time it is still possible to register new clients and the
            progressbar will only be shown after this delay..

        Notes
        -----
        If this method is called multiple times then the `stop_updater`
        method must be called the same number of times for the updater
        process to actually stop.
        """
        if self.running.is_set() is False:
            # self._update_process stores the process responsible to update
            # the progressbar. It may be finished anytime by calling the
            # stop_updater method. Also, it is set as a daemon process so
            # that we don't get errors if the program closes before the
            # process updating the progressbar ends (because the user
            # forgot to call the stop_updater method).
            self._update_process = multiprocessing.Process(
                name="ProgressBarUpdater",
                target=self._update_progress,
                args=(self._filename, start_delay))

            self._update_process.daemon = True

            self.running.set()
            self._update_process.start()

        self._start_updater_count += 1

    def stop_updater(self, timeout: Optional[float] = None) -> None:
        """
        Stop the process updating the progressbar.

        You should always call this function in your main process (the same
        that created the progressbar) after joining all the processes that
        update the progressbar. This guarantees that the progressbar
        updated any pending change and exited clearly.

        Parameters
        ----------
        timeout : float
            The timeout to join for the process to stop.

        Notes
        -----
        If the `start_updater` was called multiple times the process will
        only be stopped when `stop_updater` is called the same number of
        times.
        """
        self._start_updater_count -= 1
        if self._start_updater_count == 0:
            self.running.clear()
            # self._toc.value = time.time()
            assert (self._update_process is not None)
            self._update_process.join(timeout)

    # # TODO: Check if the duration property work correctly
    # @property
    # def duration(self, ):
    #     """Duration of the progress.

    #     Returns
    #     -------
    #     toc_minus_tic : float
    #         The duration passed until the progressbar reaches 100%.
    #     """
    #     # The progressbar is still running, calculate the duration since
    #     # the beginning
    #     if self.running.is_set():
    #         toc = time.time()
    #     else:
    #         toc = self._toc.value

    #     return toc - self._tic.value


class ProgressbarDistributedClientBase:
    """
    Proxy progressbar that behaves like a ProgressbarText object, but is
    actually updating a shared (with other clients) progressbar.

    The basic idea is that this proxy progressbar has the "progress" method
    similar to the standard ProgressbarText class. However, when this
    method is called it will update a value that will be read by a "server
    progressbar" object which is responsible to actually show the current
    progress.

    Parameters
    ----------
    client_id : ClientID
        The client ID.
    """
    def __init__(self, client_id: ClientID):
        """
        """
        self.client_id = client_id

    # Implement this method in a derived class
    def progress(self, count: int) -> None:
        """Updates the proxy progress bar.

        Parameters
        ----------
        count : int
            The new amount of progress.
        """


# xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
# xxxxxxxxxxxxxxx ProgressbarMultiProcessServer - START xxxxxxxxxxxxxxxxxxx
# xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
class ProgressbarMultiProcessServer(ProgressbarDistributedServerBase):
    """Class that prints a representation of the current progress of
    multiple process as text.

    While the :class:`ProgressbarText` class only tracks the progress of a
    single process, the :class:`ProgressbarMultiProcessServer` class can
    track the joint progress of multiple processes. This may be used, for
    instance, when you parallelize some task using the multiprocessing
    module.

    Using the ProgressbarMultiProcessServer class requires a little more work
    than using the ProgressbarText class, as it is described in the
    following:

     1. First you create an object of the ProgressbarMultiProcessServer class
        as usual. However, differently from the ProgressbarText class you
        don't pass the `finalcount` value to the progressbar yet.
     2. After that, for each process to be tracked, call the
        :meth:`register_client_and_get_proxy_progressbar` method passing
        the number equivalent to full progress for **that process**. This
        function returns a "proxy progressbar" that behaves like a regular
        ProgressbarText. Pass that proxy progressbar as an argument to that
        process so that it can call its "progress" method. Each process
        that calls the "progress" method of the received proxy progressbar
        will actually update the progress of the main
        ProgressbarMultiProcessServer object.
     3. Start all the processes and call the start_updater method of
        ProgressbarMultiProcessServer object so that the bar is updated by
        the different processes.
     4. After joining all the process (all work is finished) call the
        stop_updater method of the ProgressbarMultiProcessServer object.

    Parameters
    ----------
    progresschar : str
        Character used in the progressbar.
    message : str
        Message writen in the progressbar.
    sleep_time : float
        Time between progressbar updates (in seconds).
    filename : str
        If filename is None (default) then progress will be output to
        sys.stdout. If it is not None then the progress will be output
        to a file with name `filename`. This is usually useful for
        debugging and testing purposes.

    Examples
    --------

    .. code-block:: python

       import multiprocessing
       # Create a ProgressbarMultiProcessServer object
       pb = ProgressbarMultiProcessServer(message="some message")
       # Creates a proxy progressbar for one process passing the value
       # corresponding to 100% progress for the first process
       proxybar1 = pb.register_client_and_get_proxy_progressbar(60)
       # Creates a proxy progressbar for another process
       proxybar2 = pb.register_client_and_get_proxy_progressbar(80)
       # Create the first process passing the first proxy progressbar as
       # an argument
       p1 = multiprocessing.Process(target=some_function, args=[proxybar1])
       # Creates another process
       p2 = multiprocessing.Process(target=some_function, args=[proxybar2])
       # Start both processes
       p1.start()
       p2.start()
       # Call the start_updater method of the ProgressbarMultiProcessServer
       pb.start_updater()
       # Joint the process and then call the stop_updater method of the
       # ProgressbarMultiProcessServer
       p1.join()
       p2.join()
       pb.stop_updater()

    """
    def __init__(self,
                 progresschar: str = '*',
                 message: str = '',
                 sleep_time: float = 1.0,
                 filename: Optional[str] = None) -> None:
        """
        Initializes the ProgressbarMultiProcessServer object.

        Parameters
        ----------
        progresschar : str
            Character used in the progressbar.
        message : str
            Message writen in the progressbar.
        sleep_time : float
            Time between progressbar updates (in seconds).
        filename : str
            If filename is None (default) then progress will be output to
            sys.stdout. If it is not None then the progress will be output
            to a file with name `filename`. This is usually useful for
            debugging and testing purposes.
        """
        ProgressbarDistributedServerBase.__init__(self, progresschar, message,
                                                  sleep_time, filename)

    def _update_client_data_list(self) -> None:
        """
        This method process the communication between the client and the
        server.
        """
        # Note that since the proxybar (ProgressbarMultiProcessClient class)
        # for multiprocessing will directly modify the
        # self._client_data_list we don't need to implement a
        # _update_client_data_list method here in the
        # ProgressbarMultiProcessServer class.

    def register_client_and_get_proxy_progressbar(
            self, total_count: int) -> "ProgressbarMultiProcessClient":
        """
        Register a new "client" for the progressbar and returns a new proxy
        progressbar that the client can use to update its progress by
        calling the `progress` method of this proxy progressbar.

        The function whose process is tracked by the
        ProgressbarMultiProcessServer must must call the `progress` method of
        the returned ProgressbarMultiProcessClient object with the current
        count. This is a little less intrusive regarding the tracked
        function.

        Parameters
        ----------
        total_count : int
            Total count that will be equivalent to 100% for function.

        Returns
        -------
        obj : ProgressbarMultiProcessClient
            The proxy progressbar.
        """
        client_id = self._register_client(total_count)
        return ProgressbarMultiProcessClient(client_id, self._client_data_list)


# Used by the ProgressbarMultiProcessServer class
class ProgressbarMultiProcessClient(ProgressbarDistributedClientBase):
    """
    Proxy progressbar that behaves like a ProgressbarText object,
    but is actually updating a ProgressbarMultiProcessServer progressbar.

    The basic idea is that this proxy progressbar has the "progress" method
    similar to the standard ProgressbarText class. However, when this
    method is called it will update a value that will be read by a
    ProgressbarMultiProcessServer object instead.

    Parameters
    ----------
    client_id : ClientID
        The client ID
    client_data_list : list
        The client data list
    """
    def __init__(self, client_id: ClientID,
                 client_data_list: List[Any]) -> None:
        """Initializes the ProgressbarMultiProcessClient object."""
        ProgressbarDistributedClientBase.__init__(self, client_id)
        self._client_data_list = client_data_list

    def progress(self, count: int) -> None:
        """Updates the proxy progress bar.

        Parameters
        ----------
        count : int
            The new amount of progress.
        """
        self._client_data_list[self.client_id] = count


# xxxxxxxxxx ProgressbarMultiProcessServer - END xxxxxxxxxxxxxxxxxxxxxxxxxx


# xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
# xxxxxxxxxxxxxxx ProgressbarZMQServer xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
# xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
class ProgressbarZMQServer(ProgressbarDistributedServerBase):
    """
    Distributed "server" progressbar using ZMQ sockets.

    In order to track the progress of distributed computations two classes
    are required, one that acts as a central point and is responsible to
    actually show the progress (the server), and other class that acts as a
    proxy (the client) and is responsible to sending the current progress
    to the server. There will be one object of the "server class" and one
    or more objects of the "client class", each one tracking the progress
    of one of the distributed computations.

    This class acts like the server. It creates a ZMQ socket which expects
    (string) messages in the form "client_id:current_progress", where the
    client_id is the ID of one client progressbar previously registered
    with the register_client_and_get_proxy_progressbar method while the
    current_progress is a "number" with the current progress of that
    client.

    Note that the client proxybar for this class is implemented in the
    ProgressbarZMQClient class.

    Parameters
    ----------
    progresschar : str
        Character used in the progressbar.
    message : str
        Message writen in the progressbar.
    sleep_time : float
        Time between progressbar updates (in seconds).
    filename : str
        If filename is None (default) then progress will be output to
        sys.stdout. If it is not None then the progress will be output to a
        file with name `filename`. This is usually useful for debugging and
        testing purposes.
    ip : IPAddress
        An string representing the address of the server socket.
        Ex: '192.168.0.117', 'localhost', etc.
    port : PortNumber
        The port to bind the socket.
    """
    def __init__(self,
                 progresschar: str = '*',
                 message: str = '',
                 sleep_time: float = 1.0,
                 filename: Optional[str] = None,
                 ip: IPAddress = 'localhost',
                 port: PortNumber = 7396) -> None:
        ProgressbarDistributedServerBase.__init__(self, progresschar, message,
                                                  sleep_time, filename)

        # Create a Multiprocessing namespace
        # pylint: disable=E1101
        self._ns = self._manager.Namespace()

        # We store the IP and port of the socket in the Namespace, since
        # the socket will be created in a different process
        self._ns.ip = ip  # type: ignore
        self._ns.port = port  # type: ignore

        # This will be set to a ZMQ Context in the _update_progress method
        self._zmq_context: Optional[zmq.Context] = None
        # This will be set to a ZMQ Socket in the _update_progress method
        self._zmq_pull_socket: Optional[zmq.sugar.socket.Socket] = None

    @property
    def ip(self) -> IPAddress:
        """
        Get method for the ip property.

        Returns
        -------
        str
            The string representing the address of the server socket.
        """
        return cast(IPAddress, self._ns.ip)  # type: ignore

    @property
    def port(self) -> PortNumber:
        """
        Get method for the port property.

        Returns
        -------
        int
            The port used.
        """
        return cast(PortNumber, self._ns.port)  # type: ignore

    def register_client_and_get_proxy_progressbar(
            self, total_count: int) -> "ProgressbarZMQClient":
        """
        Register a new client progressbar and return a proxy to it.

        Parameters
        ----------
        total_count : int
            The total count for the client we are registering.

        Returns
        -------
        ProgressbarZMQClient
            The proxy progressbar.
        """
        client_id = self._register_client(total_count)
        proxybar = ProgressbarZMQClient(client_id, self.ip, self.port)
        return proxybar

    # noinspection PyUnresolvedReferences
    def _update_progress(self,
                         filename: Optional[str] = None,
                         start_delay: float = 0.0) -> None:  # pragma: no cover
        """
        Collects the progress from each registered proxy progressbar and
        updates the actual visible progressbar.

        Parameters
        ----------
        filename : str
            Name of a file where the data will be written to. If this is
            None then all progress will be printed in the standard output
            (default)
        start_delay : float
            Delay in seconds before starting the progressbar. During this
            time it is still possible to register new clients and the
            progressbar will only be shown after this delay.

        Notes
        -----
        We re-implement it here only to create the ZMQ socket. After that we
        call the base class implementation of _update_progress method. Note
        that the _update_progress method in the base class calls the
        _update_client_data_list and we indeed re-implement this method in
        this class and use the socket created here in that implementation.
        """
        # First we create the context and the socket. Then we bind the
        # socket to the respective ip:port.
        self._zmq_context = zmq.Context()
        self._zmq_pull_socket = self._zmq_context.socket(zmq.PULL)
        self._zmq_pull_socket.bind(u"tcp://*:%s" % self.port)
        ProgressbarDistributedServerBase._update_progress(self, filename)

    # This method is called in the _update_progress, which is run in a
    # different process. Therefore, the python coverage program does not
    # detect that this method is actually run (and it is) and thus we set
    # the "pragma: no cover" comment here.
    def _update_client_data_list(self) -> None:  # pragma: no cover
        """
        This method process the communication between the client and the
        server.

        This method will read the received messages in the socket which
        were sent by the clients (ProgressbarZMQClient objects) and update
        self._client_data_list variable accordingly. The messages are in
        the form "client_id:current_progress", which is parsed by the
        _parse_progress_message method.

        Notes
        -----
        This method is called inside a loop in the _update_progress method.
        """

        pending_messages = True
        while pending_messages is True and self.running.is_set():
            try:
                # Try to read a message. If this fail we will get a
                # zmq.ZMQError exception and then pending_messages will be
                # set to False so that we exit the while loop.
                assert (self._zmq_pull_socket is not None)
                message = self._zmq_pull_socket.recv_string(flags=zmq.NOBLOCK)

                # If we are here that means that a new message was
                # successfully received from the client.  Let's call the
                # _parse_progress_message method to parse the message and
                # update the self._client_data_list member variable.
                self._parse_progress_message(message)
            except zmq.ZMQError:
                pending_messages = False

    # This method run in a different process and thus the python coverage
    # program does not detect it is run even when it is run. Therefore, we
    # set the "pragma: no cover" comment here.
    def _parse_progress_message(self,
                                message: str) -> None:  # pragma: no cover
        """
        Parse the message sent from the client proxy progressbars.

        The messages sent from the proxy progressbars are in the form
        'client_id:current_count'. We need to set the element of index
        "client_id" in self._client_data_list to the value of
        "current_count". This method will simply parse the message and
        perform this operation.

        Parameters
        ----------
        message : str
            A string in the form 'client_id:current_count'.
        """
        client_id, current_count = map(int, message.split(":"))
        self._client_data_list[client_id] = current_count


# Used by the ProgressbarZMQServer class
class ProgressbarZMQClient(ProgressbarDistributedClientBase):
    """
    Proxy progressbar that behaves like a ProgressbarText object,
    but is actually updating a ProgressbarZMQServer progressbar.

    The basic idea is that this proxy progressbar has the "progress" method
    similar to the standard ProgressbarText class. However, when this
    method is called it will update a value that will be read by a
    ProgressbarZMQServer object instead.

    Parameters
    ----------
    client_id : ClientID
        The client ID.
    ip : IPAddress
        A string representing the IP address of the server.
    port : PortNumber
        The port number used by the server.
    """
    def __init__(self, client_id: ClientID, ip: IPAddress,
                 port: PortNumber) -> None:
        ProgressbarDistributedClientBase.__init__(self, client_id)
        self.ip = ip
        self.port = port

        # Function that will be called to update the progress. This
        # variable is initially set to the "_connect_and_update_progress"
        # method that will create the socket, connect it to the main
        # progressbar and finally set "_progress_func" to the "_progress"
        # method that will actually update the progress.
        self._progress_func = \
            ProgressbarZMQClient._connect_and_update_progress

        # ZMQ Variables: These variables will be set the first time the
        # progress method is called.
        self._zmq_context: Optional[zmq.Context] = None
        self._zmq_push_socket: Optional[zmq.sugar.socket.Socket] = None

    def progress(self, count: int) -> None:
        """Updates the proxy progress bar.

        Parameters
        ----------
        count : int
            The new amount of progress.
        """
        # noinspection PyArgumentList,PyTypeChecker
        self._progress_func(self, count)

    def __call__(self, count: int) -> None:
        """
        Updates the proxy progress bar.

        This method is the same as the :meth:`progress`. It is define so
        that a ProgressbarZMQClient object can behave like a function.

        Parameters
        ----------
        count : int
            The new amount of progress.
        """
        # noinspection PyArgumentList,PyTypeChecker
        self._progress_func(self, count)

    def _progress(self, count: int) -> None:
        """

        Parameters
        ----------
        count : int
            The new amount of progress.
        """
        # The message is a string composed of the client ID and the current
        # count
        message = "{0}:{1}".format(self.client_id, count)
        assert (self._zmq_push_socket is not None)
        self._zmq_push_socket.send_string(message, flags=zmq.NOBLOCK)

    def _connect_and_update_progress(self, count: int) -> None:
        """
        Creates the "push socket", connects it to the socket of the main
        progressbar and then updates the progress.

        This function will be called only in the first time the "progress"
        method is called. Subsequent calls to "progress" will actually
        calls the "_progress" method.

        Parameters
        ----------
        count : int
            The new amount of progress.
        """
        self._zmq_context = zmq.Context()
        self._zmq_push_socket = self._zmq_context.socket(zmq.PUSH)
        self._zmq_push_socket.connect("tcp://{0}:{1}".format(
            self.ip, self.port))

        # The default LINGER value for a ZMQ socket is -1, which means
        # "wait forever". That means that if the message was not received
        # by the server (the main progressbar) the process with the
        # push_socket will hang. Since we don't want that, we set the
        # LINGER option to 0 so that it does not wait for the message to be
        # received.
        self._zmq_push_socket.setsockopt(zmq.LINGER, 0)

        self._progress_func = ProgressbarZMQClient._progress
        # noinspection PyArgumentList,PyTypeChecker
        self._progress_func(self, count)


# xxxxxxxxxx ProgressbarZMQServer - END xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
